import torch
import torch.nn as nn
from reg_hnt.reghnt.optimizer import BertAdam as Adam
from reg_hnt.reghnt.util import AverageMeter
from tqdm import tqdm
SCALE = ["", "thousand", "million", "billion", "percent"]

class RegHNTPredictModel():
    def __init__(self, args, network):
        self.args = args
        self.train_loss = AverageMeter()
        self.dev_loss = AverageMeter()
        self.step = 0
        self.updates = 0
        self.network = network

        self.mnetwork = nn.DataParallel(self.network) if args.gpu_num > 1 else self.network

        if self.args.gpu_num > 0:
            self.network.cuda()

    def avg_reset(self):
        self.train_loss.reset()
        self.dev_loss.reset()

    @torch.no_grad()
    def evaluate(self, dev_data_list, epoch):
        dev_data_list.reset()
        self.network.eval()
        for batch in tqdm(dev_data_list):
            output_dict = self.network(**batch, mode="eval", epoch=epoch)
            loss = output_dict["loss"]
            self.dev_loss.update(loss.item(), 1)
        self.network.train()

    @torch.no_grad()
    def predict(self, test_data_list):
        test_data_list.reset()
        self.network.eval()
        pred_json = {}
        for batch in tqdm(test_data_list):
            output_dict = self.network.predict(**batch, mode="eval")
            pred_answer = output_dict["answer"]
            pred_scale = output_dict["scale"]
            question_id = output_dict["question_id"]
            for i in range(len(question_id)):
                pred_json[question_id[i]] = [pred_answer[i], pred_scale[i]]
        return pred_json

    @torch.no_grad()
    def predict2(self, test_data_list):
        test_data_list.reset()
        self.network.eval()
        pred_json = {}
        for batch in tqdm(test_data_list):
            output_dict = self.network.predict2(**batch, mode="eval")
            pred_answer = output_dict["answer"]
            pred_scale = output_dict["scale"]
            question_id = output_dict["question_id"]
            for i in range(len(question_id)):
                pred_json[question_id[i]] = [pred_answer[i], pred_scale[i]]
        return pred_json

    def reset(self):
        if self.args.gpu_num > 1:
            self.mnetwork.module.reset()
        else:
            self.mnetwork.reset()

    def get_df(self):
        if self.args.gpu_num > 1:
            self.mnetwork.module.get_df()
        else:
            self.mnetwork.get_df()

    def get_metrics(self, logger=None):
        if self.args.gpu_num > 1:
            return self.mnetwork.module.get_metrics(logger, True)
        else:
            return self.mnetwork.get_metrics(logger, True)


class RegHNTFineTuningModel():
    def __init__(self, args, network, state_dict=None, num_train_steps=1):
        self.args = args
        self.train_loss = AverageMeter()
        self.dev_loss = AverageMeter()
        self.step = 0
        self.updates = 0
        self.network = network
        if state_dict is not None:
            print("Load Model!")
            self.network.load_state_dict(state_dict["state"])
        self.mnetwork = nn.DataParallel(self.network) if args.gpu_num > 1 else self.network

        self.total_param = sum([p.nelement() for p in self.network.parameters() if p.requires_grad])
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_parameters = [
            {'params': [p for n, p in self.network.encoder.named_parameters() if not any(nd in n for nd in no_decay)],
             'weight_decay': args.bert_weight_decay, 'lr': args.bert_learning_rate},
            {'params': [p for n, p in self.network.encoder.named_parameters() if any(nd in n for nd in no_decay)],
             'weight_decay': 0.0, 'lr': args.bert_learning_rate},
            {'params': [p for n, p in self.network.named_parameters() if not n.startswith("encoder.")],
             "weight_decay": args.weight_decay, "lr": args.learning_rate}
        ]
        self.optimizer = Adam(optimizer_parameters,
                              lr=args.learning_rate,
                              warmup=args.warmup,
                              t_total=num_train_steps,
                              max_grad_norm=args.grad_clipping,
                              schedule=args.warmup_schedule)
        if self.args.gpu_num > 0:
            self.network.cuda()

    def avg_reset(self):
        self.train_loss.reset()
        self.dev_loss.reset()

    def update(self, tasks):
        self.network.train()
        output_dict = self.mnetwork(**tasks)
        loss = output_dict["loss"]
        loss = torch.mean(loss)
        self.train_loss.update(loss.item(), 1)
        if self.args.gradient_accumulation_steps > 1:
            loss /= self.args.gradient_accumulation_steps
        loss.backward()
        if (self.step + 1) % self.args.gradient_accumulation_steps == 0:
            self.optimizer.step()
            self.optimizer.zero_grad()
            self.updates += 1
        self.step += 1

    @torch.no_grad()
    def evaluate(self, dev_data_list, epoch):
        dev_data_list.reset()
        self.network.eval()
        for batch in dev_data_list:
            output_dict = self.network(**batch, mode="eval", epoch=epoch)
            loss = output_dict["loss"]
            self.dev_loss.update(loss.item(), 1)
        self.network.train()

    @torch.no_grad()
    def predict(self, test_data_list):
        test_data_list.reset()
        self.network.eval()
        # pred_json = {}
        for batch in tqdm(test_data_list):
            self.network.predict(**batch, mode="eval")

    def reset(self):
        if self.args.gpu_num > 1:
            self.mnetwork.module.reset()
        else:
            self.mnetwork.reset()

    def get_df(self):
        if self.args.gpu_num > 1:
            self.mnetwork.module.get_df()
        else:
            self.mnetwork.get_df()

    def get_metrics(self, logger=None):
        if self.args.gpu_num > 1:
            return self.mnetwork.module.get_metrics(logger, True)
        else:
            return self.mnetwork.get_metrics(logger, True)

    def save(self, prefix, epoch):
        network_state = dict([(k, v.cpu()) for k, v in self.network.state_dict().items()])
        other_params = {
            'optimizer': self.optimizer.state_dict(),
            'config': self.args,
            'epoch': epoch
        }
        state_path = prefix + ".pt"
        other_path = prefix + ".ot"
        torch.save(other_params, other_path)
        torch.save(network_state, state_path)
        print('model saved to {}'.format(prefix))


from xlsxwriter import Workbook


COLORS = {"light_yellow": "#fffbcb", "light_gray": "#ececec", "gray": "#b7b7b7"}

STYLES = {
    "data_table": {"border": 1, "border_color": COLORS["gray"]},
    "data_table_header": {"bold": True, "border": 1, "border_color": COLORS["gray"]},
    "data_table_active": {"bg_color": COLORS["light_yellow"], "border": 1, "border_color": COLORS["gray"]},
    "data_table_header_active": {
        "bold": True,
        "bg_color": COLORS["light_yellow"],
        "border": 1,
        "border_color": COLORS["gray"],
    },
    "bold": {"bold": True},
    "ann_table_header": {"bold": True, "bg_color": COLORS["light_gray"], "border": 1, "border_color": COLORS["gray"]},
    "ann_table_delim": {
        "bold": True,
        "bg_color": COLORS["light_gray"],
        "top": 1,
        "top_color": COLORS["gray"],
        "bottom": 1,
        "bottom_color": COLORS["gray"],
    },
}


def write_merged_cells(cell, row_num, col_num, merge_cells, merged_cells, worksheet, cell_format):
    value = str(cell.value)
    end_row = row_num + cell.rowspan - 1
    end_col = col_num + cell.colspan - 1

    for r in range(row_num, end_row + 1):
        for c in range(col_num, end_col + 1):
            merged_cells.add((r, c))
            if not merge_cells:  # setting the format for all cells separately
                worksheet.write(r, c, "", cell_format)

    if merge_cells:
        worksheet.merge_range(row_num, col_num, end_row, end_col, value, cell_format)
    else:
        worksheet.write(row_num, col_num, value, cell_format)

    return merged_cells


def write_html_table_to_excel(
    table,
    worksheet,
    workbook=None,
    start_row=0,
    start_col=0,
    write_table_props=False,
):

    style_objs = {k: workbook.add_format(v) for k, v in STYLES.items()}

    if write_table_props:
        worksheet.write(start_row, start_col, "properties", style_objs["bold"])
        start_row += 1

        for prop_name in table.props.keys():
            worksheet.write(start_row, start_col, prop_name, style_objs["bold"])
            worksheet.write(start_row, start_col + 1, table.props.get(prop_name, ""))
            start_row += 1

        worksheet.write(start_row, start_col, "")
        start_row += 1
        worksheet.write(start_row, start_col, "data", style_objs["bold"])
        start_row += 1

    row_num = start_row

    for row in table.cells:
        col_num = start_col
        for cell in row:
            if cell.is_dummy:
                continue
            style_key = "data_table"
            if cell.is_col_header or cell.is_row_header:
                style_key += "_header"
            # if cell.is_highlighted:
            #     style_key += "_active"
            cell_format = style_objs[style_key]
            worksheet.write(row_num, col_num, str(cell.value), cell_format)

            col_num += cell.colspan
        row_num += 1

    return row_num


def write_annotation_to_excel(tables, prop_list, ann_columns, out_file):
    """
    Write multiple tables to excel for manual annotation.
    :param tables: List[Dict]: list of dicts, where dict is all info about the table,
                               including the table as a `Table` object
    :param prop_list: List[str]: list of properties that are selected by the user
                                 to be included in the annotation
    :param ann_columns: List[str]: list of additional annotation columns to include in the annotation table
    :param out_file: str: path for output file
    :return: None, results are written in the file
    """

    workbook = Workbook(out_file)
    worksheet = workbook.add_worksheet()
    style_objs = {k: workbook.add_format(v) for k, v in STYLES.items()}

    ann_header = ["table_id"] + ann_columns  # all annotation columns
    worksheet.write_row(0, 0, ann_header + ["property_name", "property_value", "table"])
    worksheet.set_row(0, None, style_objs["ann_table_header"])

    start_row = 1
    start_col = len(ann_header)

    for table_info in tables:
        # writing table id
        worksheet.write(start_row, 0, table_info.get("table_id", ""))

        # writing properties
        props_end_row = start_row + len(prop_list)
        for i, prop_name in enumerate(prop_list):
            worksheet.write(start_row + i, start_col, prop_name, style_objs["bold"])
            worksheet.write(start_row + i, start_col + 1, table_info.get(prop_name, ""))

        # writing table
        table_end_row = write_html_table_to_excel(
            table_info["table"],
            worksheet,
            style_objs=style_objs,
            start_row=start_row,
            start_col=start_col + 2,
            merge_cells=False,
        )

        # writing delimiter
        end_row = max(props_end_row, table_end_row)
        worksheet.set_row(end_row, None, style_objs["ann_table_delim"])
        start_row = end_row + 1

    workbook.close()



from __future__ import division 
import torch
from torch.nn import functional


def sequence_mask(sequence_length, max_len=None):
    if max_len is None:
        max_len = sequence_length.data.max()
    batch_size = sequence_length.size(0)
    seq_range = torch.arange(0, max_len).long()
    seq_range_expand = seq_range.unsqueeze(0).expand(batch_size, max_len)
    if sequence_length.is_cuda:
        seq_range_expand = seq_range_expand.cuda()
    seq_length_expand = (sequence_length.unsqueeze(1).expand_as(seq_range_expand))
    return seq_range_expand < seq_length_expand


def masked_cross_entropy(logits, target, length):
    if torch.cuda.is_available():
        length = torch.LongTensor(length).cuda()
    else:
        length = torch.LongTensor(length)
    """
    Args:
        logits: A Variable containing a FloatTensor of size
            (batch, max_len, num_classes) which contains the
            unnormalized probability for each class.
        target: A Variable containing a LongTensor of size
            (batch, max_len) which contains the index of the true
            class for each corresponding step.
        length: A Variable containing a LongTensor of size (batch,)
            which contains the length of each data in a batch.
    Returns:
        loss: An average loss value masked by the length.
    """

    # logits_flat: (batch * max_len, num_classes)
    logits_flat = logits.view(-1, logits.size(-1))
    # log_probs_flat: (batch * max_len, num_classes)
    log_probs_flat = functional.log_softmax(logits_flat, dim=1)
    # target_flat: (batch * max_len, 1)
    target_flat = target.view(-1, 1)
    # losses_flat: (batch * max_len, 1)
    losses_flat = -torch.gather(log_probs_flat, dim=1, index=target_flat)

    # losses: (batch, max_len)
    losses = losses_flat.view(*target.size())
    # mask: (batch, max_len)
    mask = sequence_mask(sequence_length=length, max_len=target.size(1))
    losses = losses * mask.float()
    loss = losses.sum() / length.float().sum()
    # if loss.item() > 10:
    #     print(losses, target)
    return loss


def masked_cross_entropy_without_logit(logits, target, length):
    if torch.cuda.is_available():
        length = torch.LongTensor(length).cuda()
    else:
        length = torch.LongTensor(length)
    """
    Args:
        logits: A Variable containing a FloatTensor of size
            (batch, max_len, num_classes) which contains the
            unnormalized probability for each class.
        target: A Variable containing a LongTensor of size
            (batch, max_len) which contains the index of the true
            class for each corresponding step.
        length: A Variable containing a LongTensor of size (batch,)
            which contains the length of each data in a batch.
    Returns:
        loss: An average loss value masked by the length.
    """

    # logits_flat: (batch * max_len, num_classes)
    logits_flat = logits.view(-1, logits.size(-1))

    # log_probs_flat: (batch * max_len, num_classes)
    log_probs_flat = torch.log(logits_flat + 1e-12)

    # target_flat: (batch * max_len, 1)
    target_flat = target.view(-1, 1)
    # losses_flat: (batch * max_len, 1)
    losses_flat = -torch.gather(log_probs_flat, dim=1, index=target_flat)

    # losses: (batch, max_len)
    losses = losses_flat.view(*target.size())

    # mask: (batch, max_len)
    mask = sequence_mask(sequence_length=length, max_len=target.size(1))
    losses = losses * mask.float()
    loss = losses.sum() / length.float().sum()
    # if loss.item() > 10:
    #     print(losses, target)
    return loss




#coding=utf8
import copy, math
import torch
import torch.nn as nn
import torch.nn.utils.rnn as rnn_utils

def clones(module, N):
    "Produce N identical layers."
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])

def lens2mask(lens):
    bsize = lens.numel()
    max_len = lens.max()
    masks = torch.arange(0, max_len).type_as(lens).to(lens.device).repeat(bsize, 1).lt(lens.unsqueeze(1))
    masks.requires_grad = False
    return masks

def mask2matrix(mask):
    col_mask, row_mask = mask.unsqueeze(-1), mask.unsqueeze(-2)
    return col_mask & row_mask

def tile(x, count, dim=0):
    """
        Tiles x on dimension dim count times.
        E.g. [1, 2, 3], count=2 ==> [1, 1, 2, 2, 3, 3]
            [[1, 2], [3, 4]], count=3, dim=1 ==> [[1, 1, 1, 2, 2, 2], [3, 3, 3, 4, 4, 4]]
        Different from torch.repeat
    """
    if x is None:
        return x
    elif type(x) in [list, tuple]:
        return type(x)([tile(each, count, dim) for each in x])
    else:
        perm = list(range(len(x.size())))
        if dim != 0:
            perm[0], perm[dim] = perm[dim], perm[0]
            x = x.permute(perm).contiguous()
        out_size = list(x.size())
        out_size[0] *= count
        batch = x.size(0)
        x = x.contiguous().view(batch, -1) \
            .transpose(0, 1) \
            .repeat(count, 1) \
            .transpose(0, 1) \
            .contiguous() \
            .view(*out_size)
        if dim != 0:
            x = x.permute(perm).contiguous()
        return x

def rnn_wrapper(encoder, inputs, lens, cell='lstm'):
    """
        @args:
            encoder(nn.Module): rnn series bidirectional encoder, batch_first=True
            inputs(torch.FloatTensor): rnn inputs, [bsize x max_seq_len x in_dim]
            lens(torch.LongTensor): seq len for each sample, allow length=0, padding with 0-vector, [bsize]
        @return:
            out(torch.FloatTensor): output of encoder, bsize x max_seq_len x hidden_dim*2
            hidden_states([tuple of ]torch.FloatTensor): final hidden states, num_layers*2 x bsize x hidden_dim
    """
    # rerank according to lens and remove empty inputs
    sorted_lens, sort_key = torch.sort(lens, descending=True)
    nonzero_num, total_num = torch.sum(sorted_lens > 0).item(), sorted_lens.size(0)
    sort_key = sort_key[:nonzero_num]
    sorted_inputs = torch.index_select(inputs, dim=0, index=sort_key)
    # forward non empty inputs
    packed_inputs = rnn_utils.pack_padded_sequence(sorted_inputs, sorted_lens[:nonzero_num].tolist(), batch_first=True)
    packed_out, sorted_h = encoder(packed_inputs)  # bsize x srclen x dim
    sorted_out, _ = rnn_utils.pad_packed_sequence(packed_out, batch_first=True)
    if cell.upper() == 'LSTM':
        sorted_h, sorted_c = sorted_h
    # rerank according to sort_key
    out_shape = list(sorted_out.size())
    out_shape[0] = total_num
    out = sorted_out.new_zeros(*out_shape).scatter_(0, sort_key.unsqueeze(-1).unsqueeze(-1).repeat(1, *out_shape[1:]), sorted_out)
    h_shape = list(sorted_h.size())
    h_shape[1] = total_num
    h = sorted_h.new_zeros(*h_shape).scatter_(1, sort_key.unsqueeze(0).unsqueeze(-1).repeat(h_shape[0], 1, h_shape[-1]), sorted_h)
    if cell.upper() == 'LSTM':
        c = sorted_c.new_zeros(*h_shape).scatter_(1, sort_key.unsqueeze(0).unsqueeze(-1).repeat(h_shape[0], 1, h_shape[-1]), sorted_c)
        return out, (h.contiguous(), c.contiguous())
    return out, h.contiguous()

class MultiHeadAttention(nn.Module):

    def __init__(self, hidden_size, q_size, kv_size, output_size, num_heads=8, bias=True, feat_drop=0.2, attn_drop=0.0):
        super(MultiHeadAttention, self).__init__()
        self.num_heads = int(num_heads)
        self.hidden_size = hidden_size
        assert self.hidden_size % self.num_heads == 0, 'Head num %d must be divided by hidden size %d' % (num_heads, hidden_size)
        self.d_k = self.hidden_size // self.num_heads
        self.feat_drop = nn.Dropout(p=feat_drop)
        self.attn_drop = nn.Dropout(p=attn_drop)
        self.W_q = nn.Linear(q_size, self.hidden_size, bias=bias)
        self.W_k = nn.Linear(kv_size, self.hidden_size, bias=False)
        self.W_v = nn.Linear(kv_size, self.hidden_size, bias=False)
        self.W_o = nn.Linear(self.hidden_size, output_size, bias=bias)

    def forward(self, hiddens, query_hiddens, mask=None):
        ''' @params:
                hiddens : encoded sequence representations, bsize x seqlen x hidden_size
                query_hiddens : bsize [x tgtlen ]x hidden_size
                mask : length mask for hiddens, ByteTensor, bsize x seqlen
            @return:
                context : bsize x[ tgtlen x] hidden_size
        '''
        remove_flag = False
        if query_hiddens.dim() == 2:
            query_hiddens, remove_flag = query_hiddens.unsqueeze(1), True
        Q, K, V = self.W_q(self.feat_drop(query_hiddens)), self.W_k(self.feat_drop(hiddens)), self.W_v(self.feat_drop(hiddens))
        Q, K, V = Q.reshape(-1, Q.size(1), 1, self.num_heads, self.d_k), K.reshape(-1, 1, K.size(1), self.num_heads, self.d_k), V.reshape(-1, 1, V.size(1), self.num_heads, self.d_k)
        e = (Q * K).sum(-1) / math.sqrt(self.d_k) # bsize x tgtlen x seqlen x num_heads
        if mask is not None:
            e = e + ((1 - mask.float()) * (-1e20)).unsqueeze(1).unsqueeze(-1)
        a = torch.softmax(e, dim=2)
        concat = (a.unsqueeze(-1) * V).sum(dim=2).reshape(-1, query_hiddens.size(1), self.hidden_size)
        context = self.W_o(concat)
        if remove_flag:
            return context.squeeze(dim=1), a.mean(dim=-1).squeeze(dim=1)
        else:
            return context, a.mean(dim=-1)

class PoolingFunction(nn.Module):
    """ Map a sequence of hidden_size dim vectors into one fixed size vector with dimension output_size """
    def __init__(self, hidden_size=256, output_size=256, bias=True, method='attentive-pooling'):
        super(PoolingFunction, self).__init__()
        assert method in ['mean-pooling', 'max-pooling', 'attentive-pooling']
        self.method = method
        if self.method == 'attentive-pooling':
            self.attn = nn.Sequential(
                nn.Linear(hidden_size, hidden_size, bias=bias),
                nn.Tanh(),
                nn.Linear(hidden_size, 1, bias=bias)
            )
        self.mapping_function = nn.Sequential(nn.Linear(hidden_size, output_size, bias=bias), nn.Tanh()) \
            if hidden_size != output_size else lambda x: x

    def forward(self, inputs, mask=None):
        """ @args:
                inputs(torch.FloatTensor): features, batch_size x seq_len x hidden_size
                mask(torch.BoolTensor): mask for inputs, batch_size x seq_len
            @return:
                outputs(torch.FloatTensor): aggregate seq_len dim for inputs, batch_size x output_size
        """
        if self.method == 'max-pooling':
            outputs = inputs.masked_fill(~ mask.unsqueeze(-1), -1e8)
            outputs = outputs.max(dim=1)[0]
        elif self.method == 'mean-pooling':
            mask_float = mask.float().unsqueeze(-1)
            outputs = (inputs * mask_float).sum(dim=1) / mask_float.sum(dim=1)
        elif self.method == 'attentive-pooling':
            e = self.attn(inputs).squeeze(-1)
            e = e + (1 - mask.float()) * (-1e20)
            a = torch.softmax(e, dim=1).unsqueeze(1)
            outputs = torch.bmm(a, inputs).squeeze(1)
        else:
            raise ValueError('[Error]: Unrecognized pooling method %s !' % (self.method))
        outputs = self.mapping_function(outputs)
        return outputs

class FFN(nn.Module):

    def __init__(self, input_size):
        super(FFN, self).__init__()
        self.input_size = input_size
        self.feedforward = nn.Sequential(
            nn.Linear(self.input_size, self.input_size * 4),
            nn.ReLU(inplace=True),
            nn.Linear(self.input_size * 4, self.input_size)
        )
        self.layernorm = nn.LayerNorm(self.input_size)

    def forward(self, inputs):
        return self.layernorm(inputs + self.feedforward(inputs))

class Registrable(object):
    """
    A class that collects all registered components,
    adapted from `common.registrable.Registrable` from AllenNLP
    """
    registered_components = dict()

    @staticmethod
    def register(name):
        def register_class(cls):
            if name in Registrable.registered_components:
                raise RuntimeError('class %s already registered' % name)

            Registrable.registered_components[name] = cls
            return cls

        return register_class

    @staticmethod
    def by_name(name):
        return Registrable.registered_components[name]

class cached_property(object):
    """ A property that is only computed once per instance and then replaces
        itself with an ordinary attribute. Deleting the attribute resets the
        property.

        Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
        """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



from typing import Dict, List, Union, Tuple
import time
# import api.api as api
import openai
from .prompt import PromptBuilder
import random
import os
import json

def parse_api_results(engine,result, result_idx_to_eid = None, verbose = False): # 转换 api 生成的结果格式
    if engine.lower() in ["llava","llava-13b","llava-7b"]:
        raise NotImplementedError
    else:
        # parse api results
        response_dict = dict()
        for idx, g in enumerate(result['choices']): # 枚举每个生成的结果
            try:
                text = g['text'] # 生成的文本
                logprob = sum(g['logprobs']['token_logprobs']) # 生成的文本的对数形式的概率, 为每个token的对数概率之和
                eid = result_idx_to_eid[idx] # 生成的文本对应的eid
                eid_pairs = response_dict.get(eid, None) # 生成的文本对应的eid的所有 text-probablity 对 (引用形式)
                if eid_pairs is None:
                    eid_pairs = []
                    response_dict[eid] = eid_pairs
                eid_pairs.append((text, logprob))

                if verbose:
                    print(text)

            except ValueError as e:
                if verbose:
                    print('---------- Error Msg --------')
                    print(e)
                    print(text)
                    print('-----------------------------')
                pass

        return response_dict


class Generator(object):
    """
    LLM generation wrapper.
    """

    def __init__(self, args, keys=None):
        self.args = args
        self.keys = keys
        self.current_key_id = 0
        self.prompt_builder = PromptBuilder(args)

    def build_few_shot_prompt_from_file(
            self,
            file_path: str,
            n_shots: int,
            qtype: str
    ):
        """
        Build few-shot prompt for generation from file.
        """

        with open(os.path.join(self.args.template_path, file_path), 'r') as f:
            lines = f.readlines()
            
        few_shot_prompt_list = []
        one_shot_prompt = ''
        last_line = None
        pre_prompt =  None # 乱序的时候忘了把这个加上, 结果效果差不多 ?
        for line in lines:
            if line == '\n' and last_line == '\n':
                few_shot_prompt_list.append(one_shot_prompt)
                one_shot_prompt = ''
            elif line == '\n' and pre_prompt == None: # 加上最开头的提示语
                pre_prompt = one_shot_prompt
                one_shot_prompt = ''
            else:
                one_shot_prompt += line
            last_line = line
        few_shot_prompt_list.append(one_shot_prompt)
        few_shot_prompt_list = few_shot_prompt_list[:n_shots]
        few_shot_prompt_list[-1] = few_shot_prompt_list[
            -1].strip()  # It is essential for prompting to remove extra '\n'
        few_shot_prompt = '\n'.join(few_shot_prompt_list)
        return pre_prompt + '\n' + few_shot_prompt

    def build_generate_prompt(
            self,
            data_item: Dict,
            qtype: str,
            cot: bool = False
    ):
        """
        Build the generate prompt
        """
        return self.prompt_builder.build_generate_prompt(
            **data_item,
            qtype=qtype, 
            cot=cot
            # only_title= True  # TEST: 生成的时候只用 title
        )

    def generate_one_pass(
            self,
            prompts: List[Tuple], # [(g_eid, prompt)]
            args: Dict,
            verbose: bool = False
    ):
        """
        Generate one pass with llm according to the generation phase.
        """
        result_idx_to_eid = []
        for p in prompts:
            result_idx_to_eid.extend([p[0]] * args['n-samples']) # 生成 sampling_n 个答案？
        prompts = [p[1] for p in prompts]

        if verbose:
            print('\n', '*' * 20, 'LLM API Call', '*' * 20)
            for prompt in prompts:
                print(prompt)
                print('\n')
            print('- - - - - - - - - - ->>')

        status, result = self._call_llm_api(
            engine = self.args.engine,
            prompt=prompts,
            max_tokens = args['max_generation_tokens'],
            temperature = self.args.temperature,
            top_p = self.args.top_p,
            n = args['n-samples'],
            stop = self.args.stop_tokens
        )

        if not status:
            return False, None
        
        result_dict = parse_api_results(self.args.engine, result , result_idx_to_eid, verbose)
        """
        result_dict = {
            'eid': [(text1, logprob1),(text2,logprob2), ...... ],
            ......
        }
        """

        return status, result_dict
        

    def _call_llm_api(
            self,
            engine: str,
            prompt: Union[str, List],
            max_tokens,
            temperature: float,
            top_p: float,
            n: int,
            stop: List[str]
    ):
        start_time = time.time()
        result = None
        while result is None:
            try:
                if engine.lower() in ["llava","llava-13b","llava-7b"]:
                    raise ValueError(f'{engine} is not supported now.')
                else:
                    key = self.keys[self.current_key_id]
                    self.current_key_id = (self.current_key_id + 1) % len(self.keys)
                    print(f"Using openai api key: {key}")
                    result = openai.Completion.create(
                        engine=engine,
                        prompt=prompt,
                        api_key=key,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        top_p=top_p,
                        n=n, # number of samplesl,default to 20 in annotate
                        stop=stop,
                        logprobs=1
                    )
                    print('Openai api inference time:', time.time() - start_time)
                    return True, result
            except openai.error.InvalidRequestError as e:
                print(e, 'Modify the request.')
                return False, None
            except Exception as e:
                print(e, 'Retry.')
                time.sleep(5)



import torch
import torch.nn as nn
import copy



class TreeNode:  # the class save the tree node
    def __init__(self, embedding, left_flag=False):
        self.embedding = embedding
        self.left_flag = left_flag


class Score(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(Score, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.attn = nn.Linear(hidden_size + input_size, hidden_size)
        self.score = nn.Linear(hidden_size, 1, bias=False)

    def forward(self, hidden, num_embeddings, num_mask=None):
        max_len = num_embeddings.size(1)
        repeat_dims = [1] * hidden.dim()
        repeat_dims[1] = max_len
        hidden = hidden.repeat(*repeat_dims)  # B x O x H
        # For each position of encoder outputs
        this_batch_size = num_embeddings.size(0)
        energy_in = torch.cat((hidden, num_embeddings), 2).view(-1, self.input_size + self.hidden_size)
        score = self.score(torch.tanh(self.attn(energy_in)))  # (B x O) x 1
        score = score.squeeze(1)
        score = score.view(this_batch_size, -1)  # B x O
        if num_mask is not None:
            num_mask=num_mask.bool()
            score = score.masked_fill_(num_mask, -1e12)
        return score


class TreeAttn(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(TreeAttn, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.attn = nn.Linear(hidden_size + input_size, hidden_size)
        self.score = nn.Linear(hidden_size, 1)

    def forward(self, hidden, encoder_outputs, seq_mask=None):
        max_len = encoder_outputs.size(0)

        repeat_dims = [1] * hidden.dim()
        repeat_dims[0] = max_len
        hidden = hidden.repeat(*repeat_dims)  # S x B x H
        this_batch_size = encoder_outputs.size(1)

        energy_in = torch.cat((hidden, encoder_outputs), 2).view(-1, self.input_size + self.hidden_size)

        score_feature = torch.tanh(self.attn(energy_in))
        attn_energies = self.score(score_feature)  # (S x B) x 1
        attn_energies = attn_energies.squeeze(1)
        attn_energies = attn_energies.view(max_len, this_batch_size).transpose(0, 1)  # B x S
        if seq_mask is not None:
            seq_mask=seq_mask.bool()
            attn_energies = attn_energies.masked_fill_(seq_mask, -1e12)
        attn_energies = nn.functional.softmax(attn_energies, dim=1)  # B x S

        return attn_energies.unsqueeze(1)



class GenerateNode(nn.Module):
    def __init__(self, hidden_size, op_nums, embedding_size, dropout=0.5):
        super(GenerateNode, self).__init__()

        self.embedding_size = embedding_size
        self.hidden_size = hidden_size

        self.embeddings = nn.Embedding(op_nums, embedding_size)
        self.em_dropout = nn.Dropout(dropout)
        self.generate_l = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)
        self.generate_r = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)
        self.generate_lg = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)
        self.generate_rg = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)

    def forward(self, node_embedding, node_label, current_context):
        node_label_ = self.embeddings(node_label)
        node_label = self.em_dropout(node_label_)
        node_embedding = node_embedding.squeeze(1)
        current_context = current_context.squeeze(1)
        node_embedding = self.em_dropout(node_embedding)
        current_context = self.em_dropout(current_context)

        l_child = torch.tanh(self.generate_l(torch.cat((node_embedding, current_context, node_label), 1)))
        l_child_g = torch.sigmoid(self.generate_lg(torch.cat((node_embedding, current_context, node_label), 1)))
        r_child = torch.tanh(self.generate_r(torch.cat((node_embedding, current_context, node_label), 1)))
        r_child_g = torch.sigmoid(self.generate_rg(torch.cat((node_embedding, current_context, node_label), 1)))
        l_child = l_child * l_child_g
        r_child = r_child * r_child_g
        return l_child, r_child, node_label_
    


class Merge(nn.Module):
    def __init__(self, hidden_size, embedding_size, dropout=0.5):
        super(Merge, self).__init__()

        self.embedding_size = embedding_size
        self.hidden_size = hidden_size

        self.em_dropout = nn.Dropout(dropout)
        self.merge = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)
        self.merge_g = nn.Linear(hidden_size * 2 + embedding_size, hidden_size)

    def forward(self, node_embedding, sub_tree_1, sub_tree_2):
        sub_tree_1 = self.em_dropout(sub_tree_1)
        sub_tree_2 = self.em_dropout(sub_tree_2)
        node_embedding = self.em_dropout(node_embedding)

        sub_tree = torch.tanh(self.merge(torch.cat((node_embedding, sub_tree_1, sub_tree_2), 1)))
        sub_tree_g = torch.sigmoid(self.merge_g(torch.cat((node_embedding, sub_tree_1, sub_tree_2), 1)))
        sub_tree = sub_tree * sub_tree_g
        return sub_tree


class Prediction(nn.Module):
    # a seq2tree decoder with Problem aware dynamic encoding

    def __init__(self, hidden_size, op_nums, input_size, dropout=0.5):
        super(Prediction, self).__init__()

        # Keep for reference
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.op_nums = op_nums

        # Define layers
        self.dropout = nn.Dropout(dropout)

        self.embedding_weight = nn.Parameter(torch.randn(1, input_size, hidden_size))

        # for Computational symbols and Generated numbers
        self.concat_l = nn.Linear(hidden_size, hidden_size)
        self.concat_r = nn.Linear(hidden_size * 2, hidden_size)
        self.concat_lg = nn.Linear(hidden_size, hidden_size)
        self.concat_rg = nn.Linear(hidden_size * 2, hidden_size)

        self.ops = nn.Linear(hidden_size * 4, op_nums)

        self.attn = TreeAttn(hidden_size, hidden_size)
        self.attn1 = TreeAttn(hidden_size, hidden_size)
        self.attn2 = TreeAttn(hidden_size, hidden_size)
        #self.attn3 = TreeAttn(hidden_size, hidden_size)
        #self.attn4 = TreeAttn(hidden_size, hidden_size)
        #self.attn5 = TreeAttn(hidden_size, hidden_size)
        self.score = Score(hidden_size * 4, hidden_size)
        self.concat_encoder_outputs = nn.Linear(hidden_size * 2, hidden_size)

    def forward(self, node_stacks, left_childs, encoder_outputs, num_pades, padding_hidden, seq_mask, mask_nums,indices,masked_index,question_mask):

        current_embeddings = []
        for st in node_stacks:
            if len(st) == 0:
                current_embeddings.append(padding_hidden)
            else:
                current_node = st[-1]
                current_embeddings.append(current_node.embedding)

        current_node_temp = []
        for l, c in zip(left_childs, current_embeddings):
            if l is None:
                c = self.dropout(c)
                g = torch.tanh(self.concat_l(c))
                t = torch.sigmoid(self.concat_lg(c))
                current_node_temp.append(g * t)
            else:
                ld = self.dropout(l)
                c = self.dropout(c)
                g = torch.tanh(self.concat_r(torch.cat((ld, c), 1)))
                t = torch.sigmoid(self.concat_rg(torch.cat((ld, c), 1)))
                current_node_temp.append(g * t)

        current_node = torch.stack(current_node_temp)

        current_embeddings = self.dropout(current_node)

        current_attn = self.attn(current_embeddings.transpose(0, 1), encoder_outputs, seq_mask)
        current_context = current_attn.bmm(encoder_outputs.transpose(0, 1))  # B x 1 x N
        
        sen_attn1=self.attn1(current_embeddings.transpose(0, 1), encoder_outputs, question_mask[:,:,0])
        question_context1=sen_attn1.bmm(encoder_outputs.transpose(0, 1))  # B x 1 x N
        
        sen_attn2=self.attn2(current_embeddings.transpose(0, 1), encoder_outputs, question_mask[:,:,1])
        question_context2=sen_attn2.bmm(encoder_outputs.transpose(0, 1))  # B x 1 x N

        # the information to get the current quantity
        batch_size = current_embeddings.size(0)
        # predict the output (this node corresponding to output(number or operator)) with PADE

        repeat_dims = [1] * self.embedding_weight.dim()
        repeat_dims[0] = batch_size
        embedding_weight = self.embedding_weight.repeat(*repeat_dims)  # B x input_size x N
        embedding_weight = torch.cat((embedding_weight, num_pades), dim=1)  # B x O x N

        leaf_input = torch.cat((current_node, current_context,question_context1,question_context2), 2)
        leaf_input = leaf_input.squeeze(1)
        leaf_input = self.dropout(leaf_input)

        embedding_weight_ = self.dropout(embedding_weight)
        num_score = self.score(leaf_input.unsqueeze(1), embedding_weight_, mask_nums)

        op = self.ops(leaf_input)

        return num_score, op, current_node, current_context, embedding_weight,current_attn
    
    
def generate_tree_input(target, decoder_output, nums_stack_batch, num_start):
    
    # when the decoder input is copied num but the num has two pos, chose the max
    target_input = copy.deepcopy(target)
    for i in range(len(target)):
        if target_input[i] >= num_start:
            target_input[i] = 0
    return torch.LongTensor(target).cuda(), torch.LongTensor(target_input).cuda()


def copy_list(l):
    r = []
    if len(l) == 0:
        return r
    for i in l:
        if type(i) is list:
            r.append(copy_list(i))
        else:
            r.append(i)
    return r


class TreeBeam:  # the class save the beam node
    def __init__(self, score, node_stack, embedding_stack, left_childs, out):
        self.score = score
        self.embedding_stack = copy_list(embedding_stack)
        self.node_stack = copy_list(node_stack)
        self.left_childs = copy_list(left_childs)
        self.out = copy.deepcopy(out)


class TreeEmbedding:  # the class save the tree
    def __init__(self, embedding, terminal=False):
        self.embedding = embedding
        self.terminal = terminal


import datasets
import os
import json


def load_json(path):
    with open(path, 'r') as f:
        data = json.load(f)
    return data


def load_data(path_dict):
    if 'train' in path_dict:
        train_data = load_json(path_dict['train'])
    if 'dev' in path_dict:
        dev_data = load_json(path_dict['dev'])
    if 'test' in path_dict:
        test_data = load_json(path_dict['test'])
    return {'train':train_data, 'dev':dev_data, 'test':test_data}



def dataset_download(dataset_name):
    root_path = '../cache/dataset'
    if dataset_name == 'hybridqa':
        dataset_path = os.path.join(root_path, dataset_name)
        if (not os.path.exists(dataset_path)) or len(os.listdir(dataset_path)) == 0:
            # 需要写下载数据集的代码
            dataset_path = '/home/lfy/beiyongdata/HybridQA'
            pass
        else:
            pass
        train_path = os.path.join(dataset_path, "train.p.json")
        dev_path = os.path.join(dataset_path, "dev.p.json")
        test_path = os.path.join(dataset_path, "test.p.json")
    elif dataset_name == 'finqa':
        pass    
    return {"train": train_path, "dev": dev_path, "test": test_path}


import logging
import sys
import math

def create_logger(name, silent=False, to_disk=True, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log




    
    


        


import numpy as np
from tatqa_utils import to_number, is_number
import re
from enum import IntEnum

# "SPAN-TABLE-TEXT": 2 has no data
OPERATOR_CLASSES_ = {"SPAN-TEXT": 0, "SPAN-TABLE": 1, "MULTI_SPAN": 2, "CHANGE_RATIO": 3,
                    "AVERAGE": 4, "COUNT": 5, "SUM": 6, "DIFF": 7, "TIMES": 8, "DIVIDE": 9}


OPERATOR = ['+', '-', '*', '/']

SCALE = ["", "thousand", "million", "billion", "percent"]

def get_operators(derivation:str):
    res = []
    for c in derivation:
        if c in OPERATOR:
            res.append(c)
    return res

def get_operands(derivation):
    num_strs = re.split('\+|-|\*|/', derivatfion)
    result = []
    for it in num_strs:
        one = to_number(it)
        if one is not None:
            result.append(one)
    return result

def facts_to_nums(facts):
    return [to_number(f) for f in facts]

def _is_average(num_facts:list, answer):
    return round(np.average(num_facts), 2) == round(answer, 2)

def _is_change_ratio(num_facts:list, answer):
    if len(num_facts) != 2:
        return False
    cands = []
    if num_facts[1] != 0:
        ori_percent = round(100 * (num_facts[0] - num_facts[1]) / num_facts[1], 2)
        cands.append(ori_percent)
    if num_facts[0] != 0:
        ori_percent = round(100 * (num_facts[1] - num_facts[0]) / num_facts[0], 2)
        cands.append(ori_percent)
    return round(answer, 2) in cands 

def _is_division(num_facts:list, answer):
    if len(num_facts) != 2:
        return False
    cands = []
    if num_facts[1] != 0:
        cands.append(round(num_facts[0]/num_facts[1], 2))
        cands.append(100 * round(num_facts[0]/num_facts[1], 2))
    if num_facts[0] != 0:
        cands.append(round(num_facts[1]/num_facts[0], 2))
        cands.append(100 * round(num_facts[1]/num_facts[0], 2))
    return round(answer, 2) in cands

def _is_diff(num_facts:list, answer):
    if len(num_facts) != 2:
        return False
    ans_1 = round(num_facts[0] - num_facts[1], 2)
    ans_2 = round(num_facts[1] - num_facts[0], 2)
    return round(answer, 2) in (ans_1, ans_2)

def _is_sum(num_facts:list, answer):
    return round(np.sum(num_facts), 2) == round(answer, 2)

def _is_times(num_facts:list, answer):
    return round(np.prod(num_facts), 2) == round(answer, 2)

def get_operator_class(derivation:str, answer_type:str, facts:list, answer, mapping:dict, scale, OPERATOR_CLASSES):
    operator_class = None
    if answer_type == "span":
        if "table" in mapping:
            operator_class = OPERATOR_CLASSES["SPAN-TABLE"]
        else:
            operator_class = OPERATOR_CLASSES["SPAN-TEXT"]
    elif answer_type == "multi-span":
        operator_class = OPERATOR_CLASSES["MULTI_SPAN"]
    elif answer_type == "count":
        operator_class = OPERATOR_CLASSES["COUNT"]
    elif answer_type == "arithmetic":
        num_facts = facts_to_nums(facts)
        if not is_number(str(answer)):
            return None  # not support date
        if _is_change_ratio(num_facts, answer):
            operator_class = OPERATOR_CLASSES["CHANGE_RATIO"]
        elif _is_average(num_facts, answer):
            operator_class = OPERATOR_CLASSES["AVERAGE"]
        elif _is_sum(num_facts, answer):
            operator_class = OPERATOR_CLASSES["SUM"]
        elif _is_times(num_facts, answer):
            operator_class = OPERATOR_CLASSES["TIMES"]
        elif _is_diff(num_facts, answer):
            operator_class = OPERATOR_CLASSES["DIFF"]
        elif _is_division(num_facts, answer):
            operator_class = OPERATOR_CLASSES["DIVIDE"]

    return operator_class


import torch
import os
import math
from transformers import get_linear_schedule_with_warmup
import argparse
import random
import numpy as np
from torch import nn
from torch.nn import functional as F
from torch.optim import AdamW
import logging
from transformers import AutoTokenizer
from tqdm import tqdm
from torch.utils.data import DataLoader
import sys
from time import gmtime, strftime
import json
from .utils import get_training_args
from .dataset import RetrieverDataset,collate
from .model import RetrieverModel

device = torch.device("cuda")
ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")
MAX_LEN = 512


def create_logger(name, silent=False, to_disk=True, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log


class RetrieverTrainer:

    def __init__(self, caption_file_name = 'mmqa_captions_llava.json') -> None:
        args = get_training_args()
        self.args = args
        if not os.path.exists(args.output_dir):
            os.mkdir(args.output_dir)
        self.model = RetrieverModel(bert_model=args.bert_model, dropout=args.dropout) 
        self.tokenizer = AutoTokenizer.from_pretrained(args.bert_model)
        self.train_set = None
        self.dev_set = None
        self.test_set = None
        self.caption_file_name = caption_file_name
        self.loss_fn = nn.CrossEntropyLoss() if args.loss_fn_type == 'CE' else nn.BCEWithLogitsLoss(reduction='sum')

    def train(self, resume_from = None):
        logger = create_logger("Training", log_file=os.path.join(self.args.output_dir, self.args.log_file))

        if self.train_set == None:
            self.train_set = RetrieverDataset(tokenizer=self.tokenizer, data_path=self.args.data_path,
                                         split='train', caption_file_name=self.caption_file_name)
        if self.dev_set == None:
            self.dev_set = RetrieverDataset(tokenizer=self.tokenizer, data_path=self.args.data_path, 
                                       split='dev', caption_file_name=self.caption_file_name)
        
        train_loader = DataLoader(self.train_set, batch_size=self.args.per_device_train_batch_size, 
                                  collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length, self.args.image_or_text),shuffle=True)
        dev_loader = DataLoader(self.dev_set, batch_size=self.args.per_device_eval_batch_size, 
                                collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length, self.args.image_or_text))
        
        epochs = self.args.num_train_epochs
        device = torch.device("cuda")
        model = self.model
        model.to(device)
        if self.args.n_gpu > 1:
            model = torch.nn.DataParallel(model)
        # optimizer = AdamW(model.parameters(), lr=args.lr, eps=args.adam_epsilon)
        optimizer = AdamW(model.parameters(), lr=self.args.learning_rate)

        t_total = len(self.train_set) * epochs

        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=self.args.warmup_steps * t_total, num_training_steps=t_total
        )

        # outputs/retriever.pt 是否存在？
        # load model
        best_recall = 0
        model_load_path = self.args.resume_path if resume_from is None else resume_from
        if model_load_path is not None and os.path.exists(model_load_path):
            logger.info(f"loading trained parameters from {model_load_path}")
            model.load_state_dict(torch.load(model_load_path))
            logger.info("start eval....")
            recall = self.eval(dev_loader, logger)
            logger.info(f"recall: {recall}")
            best_recall = recall

        for epoch in range(epochs):
            logger.info(f"Training epoch: {epoch}")
            # try: 
            model.train()
            # averge_step = len(train_loader) // 50
            # eval_step = len(train_loader) // 8 # 1/8 个 epoch eval 一次
            logging_steps = self.args.logging_steps
            eval_steps = self.args.eval_steps
            loss_sum, step = 0, 0
            loss_func = self.loss_fn

            for i, data in enumerate(tqdm(train_loader)):
                if len(data['metadata']['docs']) > 0:
                    probs = model(data) # probs for each type
                    # if sum(data['labels'][0]).cpu().item()!=1:
                        # data['labels'] = F.softmax(probs + (1 - data['labels'].float()) * -1e20, dim=-1)

                    loss = loss_func(probs, data['labels'])
                    optimizer.zero_grad()
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
                    optimizer.step()
                    scheduler.step()
                    loss_sum += loss
                    step += 1
                if (i+1) % logging_steps == 0:
                    logger.info("Training Loss [{0:.5f}]".format(loss_sum/step))
                    loss_sum, step = 0, 0

                if (i+1) % eval_steps == 0:
                    best_recall = self.eval_process(dev_loader, logger, best_recall)
            # except KeyboardInterrupt:
            #     logger.info("Early stop!")
            #     break
                        
            best_recall = self.eval_process(dev_loader, logger, best_recall)

    def eval(self, loader = None, logger = None, save_type = None, ckpt_path = None):
        
        if logger == None:
            logger = create_logger("Evaling", log_file=os.path.join(self.args.output_dir, self.args.log_file))

        if loader == None:
            if self.dev_set == None:
                self.dev_set = RetrieverDataset(tokenizer=self.tokenizer, data_path=self.args.data_path, 
                                       split='dev', caption_file_name=self.caption_file_name)
            loader = DataLoader(self.dev_set, batch_size=self.args.per_device_eval_batch_size, 
                                collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length, self.args.image_or_text))
        model = self.model

        if ckpt_path!=None:
            model_load_path = ckpt_path
            logger.info(f"loading trained parameters from {model_load_path}")
            model.load_state_dict(torch.load(model_load_path))
            logger.info("start eval....")
            device = torch.device("cuda")
            model.to(device)
        else:
            logger.info("start eval....")
            device = torch.device("cuda")
            model.to(device)

        model.eval()
        total_gold = 0
        recall_num = 0
        scores_dict = dict()
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                if len(data['metadata']['docs']) > 0:
                    probs = model(data)
                    metadata = data['metadata']
                
                    pb_n = probs.cpu().numpy()
                    if save_type is not None:
                        scores_dict[str(i)] = dict()
                        for idx, doc_id in enumerate(metadata['docs']):
                            scores_dict[str(i)][doc_id] = float(pb_n[idx])
    
                    # 取 top-n
                    if probs.shape[0] > self.args.top_n:
                        predicts = torch.topk(probs, self.args.top_n, dim=-1).indices.cpu().numpy() # indices of preds
                    else:
                        predicts = torch.topk(probs, probs.shape[0], dim=-1).indices.cpu().numpy()
                    
                    retrieve_ids = [metadata['docs'][predicts[i]] for i in range(len(predicts))] # id of retrieved
                    targets = [metadata['gold_docs'][i] for i in range(len(metadata['gold_docs']))] # id of gold
                    total_gold += len(targets)
                    recall_num += sum([1 if targets[i] in retrieve_ids else 0 for i in range(len(targets))])
                else:
                    if save_type is not None:
                        scores_dict[str(i)] = dict()
    
        if save_type is not None:
            save_path = f'eval_result/retrieve_{save_type}.json'
            with open(save_path, 'w') as f:
                json.dump(scores_dict, f)

        logger.info(f"Total: {total_gold}\nRecall: {recall_num}")
        return recall_num / total_gold
    
    
    def eval_process(self, dev_loader, logger, best_recall):
        recall = self.eval(dev_loader, logger)
        model = self.model
        if recall > best_recall:
            best_recall = recall
            logger.info(f"best recall: {recall}")
            model_to_save = model.module if hasattr(model, "module") else model
            model_save_path = os.path.join(self.args.output_dir, f'{self.args.ckpt_file}_{self.args.image_or_text}.pt')
            logger.info(f"saving model...to {model_save_path}")
            torch.save(model_to_save.state_dict(), model_save_path)
        return best_recall

    def test(self, ckpt_path = None,save_type = None, per_device_test_batch_size = 1):
        if ckpt_path == None and self.args.ckpt_for_test==None:
            raise ValueError("ckpt_path is empty")
    
        logger = create_logger("Testing", log_file=os.path.join(self.args.output_dir, self.args.log_file))
        if self.test_set == None:
            self.test_set = RetrieverDataset(tokenizer=self.tokenizer, data_path=self.args.data_path, 
                                       split='test', caption_file_name=self.caption_file_name)
        loader = DataLoader(self.test_set, batch_size=per_device_test_batch_size, 
                                 collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length, image_or_text=self.args.image_or_text, test=True))
        model = self.model
        model_load_path = ckpt_path if ckpt_path is not None else self.args.ckpt_for_test
        logger.info(f"loading trained parameters from {model_load_path}")
        model.load_state_dict(torch.load(model_load_path))
        logger.info("start infer....")

        device = torch.device("cuda")
        model.to(device)
        model.eval()
        scores_dict = dict()
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                if len(data['metadata']['docs']) > 0:
                    probs = model(data)
                    metadata = data['metadata']
                    pb_n = probs.cpu().numpy()
                    if save_type is not None:
                        scores_dict[str(i)] = dict()
                        for idx, doc_id in enumerate(metadata['docs']):
                            scores_dict[str(i)][doc_id] = float(pb_n[idx])
                else:
                    if save_type is not None:
                        scores_dict[str(i)] = dict()

        with open(self.args.test_out_path, 'w') as f:
            json.dump(scores_dict, f, indent=4)


from torch.utils.data import Dataset, DataLoader
from transformers import AutoTokenizer
from tqdm import tqdm
import json


class RetrievalDataset(Dataset):
    def __init__(self, args, data):
        super(RetrievalDataset, self).__init__()
        tokenizer = AutoTokenizer.from_pretrained(args.plm)
        self.tokenizer = tokenizer
        self.total_data = []
        for item in tqdm(data): # 看字典的元素，xxx.keys()
            
            if item['labels'].count(1) == 0:
                continue
            
            question = item['question']
            path = '/home/lfy/UMQM/Data/HybridQA/WikiTables-WithLinks'
            table_id = item['table_id']
            
            
            with open(f'{path}/tables_tok/{table_id}.json', 'r') as f:
                table = json.load(f) 
            
            with open('{}/request_tok/{}.json'.format(path, table_id), 'r') as f:
                requested_document = json.load(f)
            
            item_data = {}
            input_ids = [] 
            question_ids = tokenizer.encode(question)
            template = '{} is {}. '
            headers =  [aa[0] for aa in table['header']]
            
            for i, row in enumerate(table['data']):
                 
                links = []
                row_ids = []
                for j, cell in enumerate(row):
                    
                    cell_string = template.format(headers[j], cell[0]) #cell[1]代表超链接
                    cell_tokens = tokenizer.tokenize(cell_string)
                    cell_ids = tokenizer.convert_tokens_to_ids(cell_tokens)
                    row_ids.extend(cell_ids)
                    row_ids += [tokenizer.sep_token_id] #加分隔符sep
                    
                    links.extend(cell[1])
                
                passage_ids = []
                for link in links:
                    
                    link_tokens = tokenizer.tokenize(requested_document[link]) #超链接
                    link_ids = tokenizer.convert_tokens_to_ids(link_tokens)
                    passage_ids.extend(link_ids)
                    passage_ids += [tokenizer.sep_token_id]
                
                
                row_ids = question_ids + row_ids + passage_ids
                input_ids.append(row_ids)
            
            item_data['input_ids'] = input_ids
            item_data['labels'] = item['labels']
            item_data['metadata'] = item #原数据
            self.total_data.append(item_data) # 标签加input
            
    def __len__(self):
        return len(self.total_data)
    def __getitem__(self, index):
        data = self.total_data[index]
        return data
    
    

class ColumnRetrievalDataset(Dataset):
    def __init__(self, args, data):
        super(ColumnRetrievalDataset, self).__init__()
        tokenizer = AutoTokenizer.from_pretrained(args.plm)
        self.tokenizer = tokenizer
        self.total_data = []
        for item in tqdm(data): 
            
            if item['labels'].count(1) == 0:
                continue
            
            question = item['question']
            path = '/home/lfy/UMQM/Data/HybridQA/WikiTables-WithLinks'
            table_id = item['table_id']
            
            
            with open(f'{path}/tables_tok/{table_id}.json', 'r') as f:
                table = json.load(f) 
            
            
            item_data = {}
            input_ids = [] 
            question_ids = tokenizer.encode(question)
            template = '{} is {}. '
            headers =  [aa[0] for aa in table['header']]
            
            header = table['header']
            labels = []
            
            data = table['data']
            
            num_columns = len(data[0])  
            num_rows = len(data)  

            for col in range(num_columns):
                column_ids = []
                column_values = [data[row][col] for row in range(num_rows)]
                for i, item_cell in enumerate(column_values):
                    cell_string = template.format(headers[col], item_cell[0])
                    cell_tokens = tokenizer.tokenize(cell_string)
                    cell_ids = tokenizer.convert_tokens_to_ids(cell_tokens)
                    column_ids += cell_ids
                    column_ids += [tokenizer.sep_token_id]
                question_column_ids = question_ids + column_ids
                input_ids.append(question_column_ids)

            item_data['input_ids'] = input_ids
            item_data['labels'] = item['labels']
            item_data['metadata'] = item 
            self.total_data.append(item_data) 
            
    def __len__(self):
        return len(self.total_data)
    def __getitem__(self, index):
        data = self.total_data[index]
        return data



# def hybridqa_content(data):
#     path = '/home/lfy/UMQM/Data/HybridQA/WikiTables-WithLinks'
#     table_id = data['table_id']
#     with open('{}/tables_tok/{}.json'.format(path, table_id), 'r') as f:
#         table = json.load(f)  
#     with open('{}/request_tok/{}.json'.format(path, table_id), 'r') as f:
#         requested_document = json.load(f)
#     content = []
#     for i, row in enumerate(table['data']):
#         # read the cell of each row, put the in a list. For special HybridQA Dataset, read the cell links.
#         cells_i = [item[0] for item in row]
#         links_iii = [item[1] for item in row]
#         links_ii = [item for sublist in links_iii for item in sublist]
#         links_i = [requested_document[link] for link in links_ii]
#         content.append((cells_i, links_i))
#     return content

# def hybridqa_header(data):
#     path = '/home/lfy/UMQM/Data/HybridQA/WikiTables-WithLinks'
#     table_id = data['table_id']
#     with open('{}/tables_tok/{}.json'.format(path, table_id), 'r') as f:
#         table = json.load(f)  
#     with open('{}/request_tok/{}.json'.format(path, table_id), 'r') as f:
#         requested_document = json.load(f)
#     header = [_[0] for _ in table['header']]
#     return header   

# def hybridqa_label(data):
#     return data['labels']


import os
import pickle
import torch
import random
import numpy as np
import dgl

class TaTQATestBatchGen(object):
    def __init__(self, args, data_mode, encoder='roberta'):
        dpath =  f"reghnt_{encoder}_cached_{data_mode}.pkl"
        self.is_train = data_mode == "train"
        self.args = args
        with open(os.path.join(args.data_dir, dpath), 'rb') as f:
            print("Load data from {}.".format(dpath))
            data = pickle.load(f)

        all_data = []
        for item in data:

            input_ids = torch.from_numpy(item["input_ids"])
            attention_mask = torch.from_numpy(item["attention_mask"])
            token_type_ids = torch.from_numpy(item["token_type_ids"])


            question_tokens = item['question_tokens']
            table = item['table']
            table_cell_type_ids = torch.from_numpy(item['table_cell_type_ids'])
            table_is_scale_type = torch.from_numpy(item['table_is_scale_type'])
            table_number = item['table_number']
            table_number_index = torch.from_numpy(item['table_number_index'])
            table_cell_index = torch.from_numpy(item['table_cell_index'])
            table_tokens = item['table_tokens']
            paragraph_tokens = item['paragraph_tokens']
            paragraph_index = torch.from_numpy(item['paragraph_index'])
            paragraph_number = item['paragraph_number']
            paragraph_number_index = torch.from_numpy(item['paragraph_number_index'])


            question_id = item['question_id']
            
            word_mask = torch.tensor(item['word_mask']) 
            number_mask = torch.tensor(item['number_mask']) 
            subword_lens = item['subword_lens']
            item_length = item['item_length']
            
            word_word_mask = item['word_word_mask']
            number_word_mask = item['number_word_mask']
            
            table_word_tokens = item['table_word_tokens']
            
            
            table_word_tokens_number = item['table_word_tokens_number']
            
            table_number_scale = torch.tensor(item['table_number_scale'])
            paragraph_number_scale = torch.tensor(item['paragraph_number_scale'])
            
            
            graph = item['graph']
            
            questions = item['question']

            question_number = item['question_number']
            question_number_index = torch.from_numpy(np.array(item['question_number_index']))
            
            
            all_data.append((input_ids, attention_mask, token_type_ids,
                question_tokens,
                table, table_cell_type_ids, table_is_scale_type, table_number, table_number_index,
                table_cell_index, table_tokens,paragraph_tokens, paragraph_index,
                paragraph_number,paragraph_number_index, question_id,
                word_mask, number_mask, subword_lens, item_length,
                word_word_mask, number_word_mask, table_word_tokens, table_word_tokens_number,
                table_number_scale, paragraph_number_scale, graph, questions,
                question_number, question_number_index))
        print("Load data size {}.".format(len(all_data)))
        self.data = TaTQATestBatchGen.make_batches(all_data, args.batch_size if self.is_train else args.eval_batch_size,
                                              self.is_train)
        self.offset = 0

    @staticmethod
    def make_batches(data, batch_size=32, is_train=True):
        if is_train:
            random.shuffle(data)
        if is_train:
            return [
                data[i: i + batch_size] if i + batch_size < len(data) else data[i:] + data[:i + batch_size - len(data)]
                for i in range(0, len(data), batch_size)]
        return [data[i:i + batch_size] for i in range(0, len(data), batch_size)]

    def reset(self):
        if self.is_train:
            indices = list(range(len(self.data)))
            random.shuffle(indices)
            self.data = [self.data[i] for i in indices]
            for i in range(len(self.data)):
                random.shuffle(self.data[i])
        self.offset = 0

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        while self.offset < len(self):
            batch = self.data[self.offset]
            self.offset += 1
            input_ids_batch, attention_mask_batch,token_type_ids_batch,\
            question_tokens_batch,\
            table_batch, table_cell_type_ids_batch,table_is_scale_type_batch, table_number_batch,\
            table_number_index_batch, table_cell_index_batch, table_tokens_batch,\
            paragraph_tokens_batch,paragraph_index_batch,paragraph_number_batch,\
            paragraph_number_index_batch, question_id_batch,\
            word_mask_batch, number_mask_batch, subword_lens_batch, item_length_batch,\
            word_word_mask_batch, number_word_mask_batch, table_word_tokens_batch,\
            table_word_tokens_number_batch, table_number_scale_batch, paragraph_number_scale_batch,\
            graph_batch, question_batch,\
                question_number_batch, question_number_index_batch = zip(*batch)
            
            bs = len(batch)
            
            input_ids = torch.LongTensor(bs, 512)
            attention_mask = torch.LongTensor(bs, 512)
            token_type_ids = torch.LongTensor(bs, 512).fill_(0)
        
            
            table = []
            table_cell_type_ids = []
            table_is_scale_type = []
            table_number = []
            table_number_index = []
            paragraph_number = []
            paragraph_number_index = []
            question_number = []
            question_number_index = []
            question_id = []
            table_word_tokens_number = []
            
            table_number_scale = []
            paragraph_number_scale = []
            
            questions = []
            
            word_mask = torch.LongTensor(bs, 512)
            number_mask = torch.LongTensor(bs, 512)
            
            max_batch_len = max(list(map(lambda x: x['q']+x['t']+x['p'], item_length_batch)))
            

            b_mask = torch.zeros([bs, max_batch_len], dtype=bool).cuda()
            b_question_mask = torch.zeros_like(b_mask)
            b_table_mask = torch.zeros_like(b_mask)
            b_paragraph_mask = torch.zeros_like(b_mask)
            b_number_mask = torch.zeros_like(b_mask)
            
            b_tokens = []
            
            for i in range(bs):
                
                question_tokens = question_tokens_batch[i]
                table_tokens = table_tokens_batch[i]
                paragraph_tokens = paragraph_tokens_batch[i]
                
                b_tokens.append(question_tokens + table_tokens + paragraph_tokens)
                              
                
                input_ids[i] = input_ids_batch[i]
                attention_mask[i] = attention_mask_batch[i]
                token_type_ids[i] = token_type_ids_batch[i]

                
                table.append(table_batch[i])
                table_cell_type_ids.append(table_cell_type_ids_batch[i])
                table_is_scale_type.append(table_is_scale_type_batch[i])
                question_number.append(question_number_batch[i])
                question_number_index.append(question_number_index_batch[i])
                table_number.append(table_number_batch[i])
                table_number_index.append(table_number_index_batch[i])
                paragraph_number.append(paragraph_number_batch[i])
                paragraph_number_index.append(paragraph_number_index_batch[i])

                question_id.append(question_id_batch[i])
                word_mask[i] = word_mask_batch[i]
                number_mask[i] = number_mask_batch[i]

                table_number_scale.append(table_number_scale_batch[i])
                paragraph_number_scale.append(paragraph_number_scale_batch[i])
                questions.append(question_batch[i])
                
                
                item_l = item_length_batch[i]
                leng = item_l['q'] + item_l['t'] + item_l['p'] 
                b_mask[i][:leng] = 1
    
                b_question_mask[i][:item_l['q']] = 1
                b_table_mask[i][item_l['q']:item_l['q']+item_l['t']] = 1
                b_paragraph_mask[i][item_l['q']+item_l['t']:item_l['q']+item_l['t']+item_l['p']]=1
                
                question_number_index_i = question_number_index_batch[i]
                table_number_index_i = table_number_index_batch[i]
                paragraph_number_index_i = paragraph_number_index_batch[i]
                for idx in question_number_index_i:
                    b_number_mask[i][idx.item()] = 1
                for idx in table_number_index_i:
                    b_number_mask[i][idx.item()+item_l['q']] = 1
                for idx in paragraph_number_index_i:
                    b_number_mask[i][idx.item()+item_l['q']+item_l['t']] = 1
                
                
                
            

            b_word_word_mask = torch.cat([torch.tensor(word_word_mask) for word_word_mask in word_word_mask_batch]).bool()
            b_number_word_mask = torch.cat([torch.tensor(number_word_mask) for number_word_mask in number_word_mask_batch]).bool()
            b_subword_lens = torch.cat([torch.tensor(subword_lens) for subword_lens in subword_lens_batch])

            word_subword_lens = b_subword_lens.masked_select(b_word_word_mask)
            number_subword_lens = b_subword_lens.masked_select(b_number_word_mask)
            
            max_word_subword_len = max(word_subword_lens)
            max_number_subword_len = max(number_subword_lens)
            
            word_subword_mask = torch.zeros([word_subword_lens.size(0), max_word_subword_len], dtype=bool)
            number_subword_mask = torch.zeros([number_subword_lens.size(0), max_number_subword_len], dtype=bool)
            


            for i in range(len(word_subword_mask)):
                word_subword_mask[i][:word_subword_lens[i]] = 1
            for i in range(len(number_subword_mask)):
                number_subword_mask[i][:number_subword_lens[i]] = 1
            
            
            max_question_len = max([item_l['q'] for item_l in item_length_batch])
            max_table_len = max([item_l['t'] for item_l in item_length_batch])
            max_table_word_len = max([len(subword_lens_batch[i])-item_l['p'] - item_l['q'] for i, item_l in enumerate(item_length_batch)])
            max_paragraph_len = max([item_l['p'] for item_l in item_length_batch])
            

            question_lens = torch.tensor([item_l['q'] for item_l in item_length_batch]).cuda()
            table_lens = torch.tensor([item_l['t'] for item_l in item_length_batch]).cuda()
            table_word_lens = torch.tensor([len(subword_lens_batch[i])-item_l['p'] - item_l['q'] for i, item_l in enumerate(item_length_batch)]).cuda()
            paragraph_lens = torch.tensor([item_l['p'] for item_l in item_length_batch]).cuda()
            
            lens = {"question":question_lens, "table":table_lens, "table_word":table_word_lens, "paragraph":paragraph_lens}
            max_len = {"question":max_question_len,"table":max_table_len,"table_word":max_table_word_len,"paragraph":max_paragraph_len}
            mask = {"b":b_mask, "question":b_question_mask, "table":b_table_mask, "paragraph":b_paragraph_mask,
                    "number": b_number_mask}

            max_pl_len = max([len(subword_lens) for subword_lens in subword_lens_batch])

            b_pl_mask = torch.zeros([bs, max_pl_len]).bool().cuda()
            q_pl_mask = torch.zeros_like(b_pl_mask)
            t_pl_mask = torch.zeros_like(b_pl_mask)
            p_pl_mask = torch.zeros_like(b_pl_mask)
            question_pl_mask = torch.zeros([bs, max_question_len]).bool().cuda()
            table_pl_mask = torch.zeros([bs, max_table_word_len]).bool().cuda()
            paragraph_pl_mask = torch.zeros([bs, max_paragraph_len]).bool().cuda()
            for i in range(bs):
                b_pl_mask[i][:question_lens[i]+table_word_lens[i]+paragraph_lens[i]] = 1
                question_pl_mask[i][:question_lens[i]] = 1
                table_pl_mask[i][:table_word_lens[i]] = 1
                paragraph_pl_mask[i][:paragraph_lens[i]] = 1
                q_pl_mask[i][:question_lens[i]] = 1
                t_pl_mask[i][question_lens[i]:question_lens[i]+table_word_lens[i]] = 1
                p_pl_mask[i][question_lens[i]+table_word_lens[i]:question_lens[i]+table_word_lens[i]+paragraph_lens[i]] = 1     
            
            table_word_tokens_number_len = 0
            max_table_word_tokens_number_len = 0
            word_tokens_number = []
            for table_word_tokens_number in table_word_tokens_number_batch:
                table_word_tokens_number_len += len(table_word_tokens_number)
                max_table_word_tokens_number_len = max(max_table_word_tokens_number_len, max(table_word_tokens_number))
                word_tokens_number += table_word_tokens_number
            
            t_w_t_n_mask = torch.zeros([table_word_tokens_number_len, max_table_word_tokens_number_len]).bool().cuda()
            
            for i in range(len(t_w_t_n_mask)):
                t_w_t_n_mask[i][:word_tokens_number[i]] = 1
                
            b_graph = dgl.batch([g['dgl'] for g in graph_batch]).to("cuda:0")
            b_relation = torch.cat([torch.tensor(g['relations']) for g in graph_batch]).cuda()
            b_src = torch.cat([torch.tensor(g['src']) for g in graph_batch]).cuda()
            b_dst = torch.cat([torch.tensor(g['dst']) for g in graph_batch]).cuda()

            pl_mask = {"b":b_pl_mask,"question":question_pl_mask,"table":table_pl_mask,"paragraph":paragraph_pl_mask,
                       "q":q_pl_mask, "t":t_pl_mask, "p":p_pl_mask}
            
            out_batch = {"input_ids": input_ids, "attention_mask": attention_mask, "token_type_ids":token_type_ids,
            "word_subword_lens":word_subword_lens, "number_subword_lens":number_subword_lens,
            "word_subword_mask":word_subword_mask, "number_subword_mask":number_subword_mask,
            "lens": lens, "max_len": max_len, "pl_mask":pl_mask, "mask": mask,
            "table":table, "table_cell_type_ids":table_cell_type_ids,
            "table_is_scale_type":table_is_scale_type,"table_number":table_number,"table_number_index":table_number_index,
            "paragraph_number":paragraph_number,"paragraph_number_index":paragraph_number_index,
            "question_id":question_id, "word_mask":word_mask, "number_mask":number_mask,
            "b_word_word_mask":b_word_word_mask, "b_number_word_mask":b_number_word_mask, "t_w_t_n_mask":t_w_t_n_mask,
            "b_tokens":b_tokens, "b_graph": b_graph,
            "b_relation": b_relation, "b_src": b_src, "b_dst": b_dst, "t_scale":table_number_scale,"p_scale":paragraph_number_scale,
            "questions":questions,"question_number":question_number, "question_number_index":question_number_index
            }

            if self.args.cuda:
                for k in out_batch.keys():
                    if isinstance(out_batch[k], torch.Tensor):
                        out_batch[k] = out_batch[k].cuda()

            yield  out_batch


"""
General utilities.
"""
import json
import os
from typing import List, Union, Dict
from functools import cmp_to_key
import math
from collections.abc import Iterable

from datasets import load_dataset

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")

def get_type(type_str):
    if type_str in ['ImageQ', 'ImageListQ']:
        return 'image'
    elif type_str in ['TableQ']:
        return 'table'
    elif type_str in ['TextQ']:
        return 'text'
    else:
        return 'compose'

def majority_vote(
        nsqls: List,
        pred_answer_list: List,
        allow_none_and_empty_answer: bool = False,
        allow_error_answer: bool = False,
        answer_placeholder: Union[str, int] = '<error|empty>',
        vote_method: str = 'prob',
        answer_biased: Union[str, int] = None,
        answer_biased_weight: float = None,
):
    """
    Determine the final nsql execution answer by majority vote.
    """

    def _compare_answer_vote_simple(a, b):
        """
        First compare occur times. If equal, then compare max nsql logprob.
        """
        if a[1]['count'] > b[1]['count']:
            return 1
        elif a[1]['count'] < b[1]['count']:
            return -1
        else:
            if a[1]['nsqls'][0][1] > b[1]['nsqls'][0][1]:
                return 1
            elif a[1]['nsqls'][0][1] == b[1]['nsqls'][0][1]:
                return 0
            else:
                return -1

    def _compare_answer_vote_with_prob(a, b):
        """
        Compare prob sum.
        """
        return 1 if sum([math.exp(nsql[1]) for nsql in a[1]['nsqls']]) > sum(
            [math.exp(nsql[1]) for nsql in b[1]['nsqls']]) else -1

    # Vote answers
    candi_answer_dict = dict()
    for (nsql, logprob), pred_answer in zip(nsqls, pred_answer_list):
        if allow_none_and_empty_answer:
            if pred_answer == [None] or pred_answer == []:
                pred_answer = [answer_placeholder]
        if allow_error_answer:
            if pred_answer == '<error>':
                pred_answer = [answer_placeholder]

        # Invalid execution results
        if pred_answer == '<error>' or pred_answer == [None] or pred_answer == []:
            continue
        if candi_answer_dict.get(tuple(pred_answer), None) is None:
            candi_answer_dict[tuple(pred_answer)] = {
                'count': 0,
                'nsqls': []
            }
        answer_info = candi_answer_dict.get(tuple(pred_answer), None)
        answer_info['count'] += 1
        answer_info['nsqls'].append([nsql, logprob])

    # All candidates execution errors
    if len(candi_answer_dict) == 0:
        return answer_placeholder, [(nsqls[0][0], nsqls[0][-1])]

    # Sort
    if vote_method == 'simple':
        sorted_candi_answer_list = sorted(list(candi_answer_dict.items()),
                                          key=cmp_to_key(_compare_answer_vote_simple), reverse=True)
    elif vote_method == 'prob':
        sorted_candi_answer_list = sorted(list(candi_answer_dict.items()),
                                          key=cmp_to_key(_compare_answer_vote_with_prob), reverse=True)
    elif vote_method == 'answer_biased':
        # Specifically for Tabfact entailed answer, i.e., `1`.
        # If there exists nsql that produces `1`, we consider it more significant because `0` is very common.
        assert answer_biased_weight is not None and answer_biased_weight > 0
        for answer, answer_dict in candi_answer_dict.items():
            if answer == (answer_biased,):
                answer_dict['count'] *= answer_biased_weight
        sorted_candi_answer_list = sorted(list(candi_answer_dict.items()),
                                          key=cmp_to_key(_compare_answer_vote_simple), reverse=True)
    elif vote_method == 'lf_biased':
        # Assign weights to different types of logic forms (lf) to control interpretability and coverage
        for answer, answer_dict in candi_answer_dict.items():
            count = 0
            for nsql, _ in answer_dict['nsqls']:
                if 'map@' in nsql:
                    count += 10
                elif 'ans@' in nsql:
                    count += 10
                else:
                    count += 1
            answer_dict['count'] = count
        sorted_candi_answer_list = sorted(list(candi_answer_dict.items()),
                                          key=cmp_to_key(_compare_answer_vote_simple), reverse=True)
    else:
        raise ValueError(f"Vote method {vote_method} is not supported.")

    pred_answer_info = sorted_candi_answer_list[0]
    pred_answer, pred_answer_nsqls = list(pred_answer_info[0]), pred_answer_info[1]['nsqls']
    return pred_answer, pred_answer_nsqls


def load_data_split(dataset_to_load, split, data_dir=os.path.join(ROOT_DIR, 'datasets/')):
    dataset_split_loaded = load_dataset(
        path=os.path.join(data_dir, "{}.py".format(dataset_to_load)),
        cache_dir=os.path.join(data_dir, "data"),streaming=True)[split]
    if dataset_to_load == 'mmqa': # DOING
        new_dataset_split_loaded = []
        for data_item in dataset_split_loaded:
            data_item['table']['page_title'] = data_item['table']['title']
            new_dataset_split_loaded.append(data_item)
        dataset_split_loaded = new_dataset_split_loaded
    else:
        raise ValueError(f'{dataset_to_load} dataset is not supported now.')
    return dataset_split_loaded


def pprint_dict(dic):
    print(json.dumps(dic, indent=2))


def flatten(nested_list):
    for x in nested_list:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x

def parse_ans(ans_text, type, cot):
    # TODO: 增加对 python 的处理
    if not cot:
        return True, ans_text
    else:
        answer = ans_text.split("Answer: ")[1].strip()
        return True, answer


import json
import os
import sys
from abc import ABC, abstractmethod
from typing import List, Union, Dict, Optional
from transformers import HfArgumentParser
from icl import GPT, fix_seed, Logger, print_time, ICLArguments
from icl import GPTDataSet


class ICL(ABC):
    def __init__(self, model: GPT, dataset: GPTDataSet):
        self.args = HfArgumentParser(ICLArguments).parse_args_into_dataclasses()[0]
        self.model = model
        self.dataset = dataset
        fix_seed(self.args.random_seed)

    def infer(self, demo_prefix: Optional[str], cot_trigger: Optional[str], answer_trigger: str):
        print('-' * 20)
        print(self.args)
        print('-' * 20)
        if not os.path.exists(self.args.logging_dir):
            os.mkdir(self.args.logging_dir)
        sys.stdout = Logger(os.path.join(self.args.logging_dir, print_time(1) + '.log'), sys.stdout)
        print_time()
        demo_input = self.create_demo_input(self.dataset.demos, demo_prefix, cot_trigger, answer_trigger)

        self.model.set_prompt(demo_input)
        for i, one in enumerate(self.dataset.data):
            if i < self.args.resume_id - 1:
                continue
            print('-' * 20)
            print("{}st data".format(i + 1))
            print("id: {}".format(one["id"]))
            data_input = self.create_data_input(one, cot_trigger, answer_trigger, self.args.truncation)
            print("demo input: \n{}".format(demo_input))
            print("data input: \n{}".format(data_input))
            gpt_out = self.model.getResponse(
                data_input,
                self.args.temperature,
                self.args.max_length,
                self.args.api_time_interval
            )
            print("gpt output: {}".format(gpt_out))
            if answer_trigger not in gpt_out:
                answer_extract = self.model.getResponse(
                    data_input + gpt_out + answer_trigger,
                    self.args.temperature,
                    self.args.max_length,
                    self.args.api_time_interval
                )
            else:
                answer_extract = gpt_out.split(answer_trigger)[-1]
            answer = self.answer_post_proc(answer_extract)
            print("pred answer: {}".format(answer))
            self.save_prediction(self.args.output_path, one, answer)

    @staticmethod
    def remove_space(text_in):
        res = []
        for tmp in text_in.split(" "):
            if tmp != "":
                res.append(tmp)
        return " ".join(res)

    def table_flatten(self, heads: List[str], row: List[str]):
        res = ""
        if heads[0]:
            res += (heads[0] + " ")
        for head, cell in zip(heads[1:], row[1:]):
            res += ("the " + row[0] + " of " + head + " is " + cell + " ; ")
        res = self.remove_space(res)
        return res.strip()

    @staticmethod
    def table_markdown(row: List[str]):
        res = ""
        for cell in row:
            if cell != "":
                res += ("| " + cell + " ")
            else:
                res += "| - "
        return res + "|"

    @abstractmethod
    def create_demo_input(
            self,
            demos: List[Dict[str, str]],
            demo_prefix: Optional[str],
            cot_trigger: Optional[str],
            answer_trigger: str,
    ) -> Union[str, List[Dict[str, str]]]:
        """
        According to the demos from the dataset, create the demo that will be input to the openai model
        :param demo_prefix: The prefix of demo_input like "Reading the texts and tables and try your best to answer the question."
        :param demos: demos list from the dataset.create_demos()
        :param cot_trigger: the prompt to trigger cot like "Let's think step by step."
        :param answer_trigger: the prompt to trigger the answer like "Therefore, the answer is "
        :return: str for davinci like
        "Q: demo1
         A: answer1

         Q: demo2
         A: ……"

         and List[Dict] for turbo like
         [{"role": "user", "content": demo1}, {"role": "assistant", "content": answer1},
         {"role": "user", "content": demo2}, {"role": "assistant", "content": answer2},
         ……]
        """
        pass

    @abstractmethod
    def create_data_input(
            self,
            data: Dict[str, str],
            cot_trigger: Optional[str],
            answer_trigger: str,
            truncation: int
    ) -> str:
        """
        According to the data from the dataset, create the data that will be input to the openai model
        :param data: one data from the dataset
        :param cot_trigger: the prompt to trigger cot like "Let's think step by step."
        :param answer_trigger: the prompt to trigger the answer like "Therefore, the answer is "
        :param truncation: cut off the max length of content
        :return: str for all Openai model like
        "Q: paragraphs:……
        tables:……
        question:……
        A: cot_trigger
        """
        pass

    @abstractmethod
    def answer_post_proc(self, answer_extract: str) -> str:
        """
        Post-processing the answer from gpt output
        :param answer_extract: answer from gpt_output.split(answer_trigger)[-1]
        :return: processing the answer if needed.
        """
        pass

    @abstractmethod
    def save_prediction(
            self,
            output_path: str,
            data: Dict[str, str],
            answer: str
    ) -> None:
        """
        Saving the answer with the form that you are fond of
        :param output_path: your arg input
        :param data: one data from the dataset
        :param answer: the answer after post-processing.
        :return: None
        """
        pass


class turboICL(ICL):
    def create_demo_input(self, demos: List[dict], demo_prefix: Optional[str], cot_trigger: Optional[str], answer_trigger: str) -> Union[str, List[Dict[str, str]]]:
        if demo_prefix is not None:
            demo_input = [{"role": "user", "content": demo_prefix}]
        else:
            demo_input = []
        for demo in demos:
            demo_input.append({
                "role": "user",
                "content": demo["question"]
            })
            if cot_trigger is not None:
                demo_input.append({
                    "role": "assistant",
                    "content": cot_trigger + demo["rationale"] + answer_trigger + demo["answer"]
                })
            else:
                demo_input.append({
                    "role": "assistant",
                    "content": answer_trigger + demo["answer"]
                })
        return demo_input

    def create_data_input(self, data: dict, cot_trigger: Optional[str], answer_trigger: Optional[str], truncation: int) -> str:
        heads = []
        content = ""
        if data["texts"] is not None and len(data["texts"]):
            content += "paragraphs:\n"
            for text in data["texts"]:
                content += (text + "\n")
        if data["rows"] is not None and len(data["rows"]):
            content += "table:\n"
            for i, row in enumerate(data["rows"]):
                if self.args.use_table_markdown:
                    content += (self.table_markdown(row) + "\n")
                elif self.args.use_table_flatten:
                    if i == 0:
                        heads = row
                    else:
                        content += (self.table_flatten(heads, row) + "\n")
                else:
                    raise NotImplementedError
        if cot_trigger is not None:
            question = "question:\n" + data["question"] + "\n" + cot_trigger
        else:
            question = "question:\n" + data["question"]
        if len(content) > truncation - len(question):
            content = content[:truncation - len(question) - 1] + '\n'
        content += question
        return content

    def answer_post_proc(self, answer_extract: str) -> str:
        return answer_extract

    def save_prediction(self, output_path: str, data: dict, answer: str) -> None:
        if not os.path.isfile(output_path):
            with open(output_path, 'w') as file:
                json.dump([], file)
        output_data = {
            "uid": data["id"],
            "predicted_program": [],
            "predicted_ans": answer
        }
        with open(output_path, 'r+') as file:
            file_data = json.load(file)
            file_data.append(output_data)
            file.seek(0)
            json.dump(file_data, file, indent=4)


class davinciICL(ICL):
    def create_demo_input(self, demos: List[dict], demo_prefix: Optional[str], cot_trigger: Optional[str], answer_trigger: str) -> Union[str, List[Dict[str, str]]]:
        if demo_prefix is not None:
            demo_input = demo_prefix + "\n"
        else:
            demo_input = ""
        for demo in demos:
            if cot_trigger is not None:
                demo_input += ("Q: " + demo["question"] + "\n" +
                               "A: " + cot_trigger + demo["rationale"] +
                               answer_trigger + demo["answer"] + "\n\n")
            else:
                demo_input += ("Q: " + demo["question"] + "\n" +
                               "A: " + answer_trigger + demo["answer"] + "\n\n")
        return demo_input

    def create_data_input(self, data: dict, cot_trigger: Optional[str], answer_trigger: str, truncation: int) -> str:
        heads = []
        content = "Q: \n"
        if data["texts"] is not None and len(data["texts"]):
            content += "paragraphs:\n"
            for text in data["texts"]:
                content += (text + "\n")
        if data["rows"] is not None and len(data["rows"]):
            content += "table:\n"
            for i, row in enumerate(data["rows"]):
                if self.args.use_table_markdown:
                    content += (self.table_markdown(row) + "\n")
                elif self.args.use_table_flatten:
                    if i == 0:
                        heads = row
                    else:
                        content += (self.table_flatten(heads, row) + "\n")
                else:
                    raise NotImplementedError
        if cot_trigger is not None:
            question = "question:\n" + data["question"] + "\n" + "A: " + cot_trigger
        else:
            question = "question:\n" + data["question"] + "\n" + "A: "
        if len(content) > truncation - len(question):
            content = content[:truncation - len(question) - 1] + '\n'
        content += question
        return content

    def answer_post_proc(self, answer_extract: str) -> str:
        return answer_extract

    def save_prediction(self, output_path: str, data: dict, answer: str) -> None:
        if not os.path.isfile(output_path):
            with open(output_path, 'w') as file:
                json.dump([], file)
        output_data = {
            "uid": data["id"],
            "predicted_program": [],
            "predicted_ans": answer
        }
        with open(output_path, 'r+') as file:
            file_data = json.load(file)
            file_data.append(output_data)
            file.seek(0)
            json.dump(file_data, file, indent=4)



import torch

from typing import Dict, Optional, Sequence, Union

from transformers import DataCollatorWithPadding, BatchEncoding
from transformers.tokenization_utils import PreTrainedTokenizer

from .utils import IGNORE_INDEX


class DynamicDataCollatorWithPadding(DataCollatorWithPadding):
    r"""
    Inherits DataCollatorWithPadding. It is capable of dynamically padding for batched data.
    """
    def __init__(
            self,
            tokenizer: PreTrainedTokenizer,
            ignore_pad_token_for_loss: Optional[bool] = False
    ):
        super().__init__(tokenizer, padding=True)
        self.label_pad_token_id = IGNORE_INDEX if ignore_pad_token_for_loss else tokenizer.pad_token_id

    def get_attention_masks(self, input_ids: torch.Tensor, device: torch.device) -> torch.Tensor:
        r"""
        Generates attention masks for left-padded sequences.
        """
        batch_size, seq_length = input_ids.size()
        attention_mask = torch.ones((batch_size, seq_length), device=device)

        for i, seq in enumerate(input_ids):
            attention_mask[i, :(seq != self.tokenizer.pad_token_id).nonzero()[0].item()] = 0 # padding

        attention_mask = attention_mask.bool()
        return attention_mask

    def __call__(self, features: Sequence[Dict[str, Union[torch.Tensor, Sequence[int]]]]) -> BatchEncoding:
        r"""
        Pads batched data to the longest sequence in the batch.

        We adopt left-padding in both training and evaluation.
        """
        if isinstance(features[0]["input_ids"], torch.Tensor):
            input_ids = [feature["input_ids"].clone().detach().flip(0) for feature in features]
        else:
            input_ids = [torch.tensor(feature["input_ids"]).flip(0) for feature in features]

        if "labels" in features[0]:
            if isinstance(features[0]["labels"], torch.Tensor):
                labels = [feature["labels"].clone().detach().flip(0) for feature in features]
            else:
                labels = [torch.tensor(feature["labels"]).flip(0) for feature in features]
            input_ids = input_ids + labels # pad them to the same length

        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids,
            batch_first=True,
            padding_value=self.tokenizer.pad_token_id
        ).flip(-1)

        batch = {}

        if "labels" in features[0]:
            input_ids, labels = input_ids.split(len(features), dim=0)
            labels = torch.where(labels != self.tokenizer.pad_token_id, labels, self.label_pad_token_id)
            batch["labels"] = labels

        batch["input_ids"] = input_ids
        batch["attention_mask"] = self.get_attention_masks(input_ids, device=input_ids.device)

        return BatchEncoding(batch)




import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import allennlp as util

def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

def swish(x):
    return x * torch.sigmoid(x)


class ResidualGRU(nn.Module):
    def __init__(self, hidden_size, dropout=0.1, num_layers=2):
        super(ResidualGRU, self).__init__()
        self.enc_layer = nn.GRU(input_size=hidden_size, hidden_size=hidden_size // 2, num_layers=num_layers,
                                batch_first=True, dropout=dropout, bidirectional=True)
        self.enc_ln = nn.LayerNorm(hidden_size)

    def forward(self, input):
        output, _ = self.enc_layer(input)
        return self.enc_ln(output + input)


class FFNLayer(nn.Module):
    def __init__(self, input_dim, intermediate_dim, output_dim, dropout, layer_norm=True):
        super(FFNLayer, self).__init__()
        self.fc1 = nn.Linear(input_dim, intermediate_dim)
        if layer_norm:
            self.ln = nn.LayerNorm(intermediate_dim)
        else:
            self.ln = None
        self.dropout_func = nn.Dropout(dropout)
        self.fc2 = nn.Linear(intermediate_dim, output_dim)

    def forward(self, input):
        inter = self.fc1(self.dropout_func(input))
        inter_act = gelu(inter)
        if self.ln:
            inter_act = self.ln(inter_act)
        return self.fc2(inter_act)


class GCN(nn.Module):

    def __init__(self, node_dim, extra_factor_dim=0, iteration_steps=1):
        super(GCN, self).__init__()

        self.node_dim = node_dim
        self.iteration_steps = iteration_steps

        self._node_weight_fc = torch.nn.Linear(node_dim + extra_factor_dim, 1, bias=True)

        self._self_node_fc = torch.nn.Linear(node_dim, node_dim, bias=True)
        self._dd_node_fc_left = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._qq_node_fc_left = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._dq_node_fc_left = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._qd_node_fc_left = torch.nn.Linear(node_dim, node_dim, bias=False)

        self._dd_node_fc_right = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._qq_node_fc_right = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._dq_node_fc_right = torch.nn.Linear(node_dim, node_dim, bias=False)
        self._qd_node_fc_right = torch.nn.Linear(node_dim, node_dim, bias=False)

    def forward(self, d_node, q_node, d_node_mask, q_node_mask, graph, extra_factor=None):

        d_node_len = d_node.size(1)
        q_node_len = q_node.size(1)

        diagmat = torch.diagflat(torch.ones(d_node.size(1), dtype=torch.long, device=d_node.device))
        diagmat = diagmat.unsqueeze(0).expand(d_node.size(0), -1, -1)
        dd_graph = d_node_mask.unsqueeze(1) * d_node_mask.unsqueeze(-1) * (1 - diagmat)
        dd_graph_left = dd_graph * graph[:, :d_node_len, :d_node_len]
        dd_graph_right = dd_graph * (1 - graph[:, :d_node_len, :d_node_len])

        diagmat = torch.diagflat(torch.ones(q_node.size(1), dtype=torch.long, device=q_node.device))
        diagmat = diagmat.unsqueeze(0).expand(q_node.size(0), -1, -1)
        qq_graph = q_node_mask.unsqueeze(1) * q_node_mask.unsqueeze(-1) * (1 - diagmat)
        qq_graph_left = qq_graph * graph[:, d_node_len:, d_node_len:]
        qq_graph_right = qq_graph * (1 - graph[:, d_node_len:, d_node_len:])

        dq_graph = d_node_mask.unsqueeze(-1) * q_node_mask.unsqueeze(1)
        dq_graph_left = dq_graph * graph[:, :d_node_len, d_node_len:]
        dq_graph_right = dq_graph * (1 - graph[:, :d_node_len, d_node_len:])

        qd_graph = q_node_mask.unsqueeze(-1) * d_node_mask.unsqueeze(1)
        qd_graph_left = qd_graph * graph[:, d_node_len:, :d_node_len]
        qd_graph_right = qd_graph * (1 - graph[:, d_node_len:, :d_node_len])


        d_node_neighbor_num = dd_graph_left.sum(-1) + dd_graph_right.sum(-1) + dq_graph_left.sum(-1) + dq_graph_right.sum(-1)
        d_node_neighbor_num_mask = (d_node_neighbor_num >= 1).long()
        d_node_neighbor_num = util.replace_masked_values(d_node_neighbor_num.float(), d_node_neighbor_num_mask, 1)

        q_node_neighbor_num = qq_graph_left.sum(-1) + qq_graph_right.sum(-1) + qd_graph_left.sum(-1) + qd_graph_right.sum(-1)
        q_node_neighbor_num_mask = (q_node_neighbor_num >= 1).long()
        q_node_neighbor_num = util.replace_masked_values(q_node_neighbor_num.float(), q_node_neighbor_num_mask, 1)


        all_d_weight, all_q_weight = [], []
        for step in range(self.iteration_steps):
            if extra_factor is None:
                d_node_weight = torch.sigmoid(self._node_weight_fc(d_node)).squeeze(-1)
                q_node_weight = torch.sigmoid(self._node_weight_fc(q_node)).squeeze(-1)
            else:
                d_node_weight = torch.sigmoid(self._node_weight_fc(torch.cat((d_node, extra_factor), dim=-1))).squeeze(-1)
                q_node_weight = torch.sigmoid(self._node_weight_fc(torch.cat((q_node, extra_factor), dim=-1))).squeeze(-1)

            all_d_weight.append(d_node_weight)
            all_q_weight.append(q_node_weight)

            self_d_node_info = self._self_node_fc(d_node)
            self_q_node_info = self._self_node_fc(q_node)

            dd_node_info_left = self._dd_node_fc_left(d_node)
            qd_node_info_left = self._qd_node_fc_left(d_node)
            qq_node_info_left = self._qq_node_fc_left(q_node)
            dq_node_info_left = self._dq_node_fc_left(q_node)

            dd_node_weight = util.replace_masked_values(
                    d_node_weight.unsqueeze(1).expand(-1, d_node_len, -1),
                    dd_graph_left,
                    0)

            qd_node_weight = util.replace_masked_values(
                    d_node_weight.unsqueeze(1).expand(-1, q_node_len, -1),
                    qd_graph_left,
                    0)

            qq_node_weight = util.replace_masked_values(
                    q_node_weight.unsqueeze(1).expand(-1, q_node_len, -1),
                    qq_graph_left,
                    0)

            dq_node_weight = util.replace_masked_values(
                    q_node_weight.unsqueeze(1).expand(-1, d_node_len, -1),
                    dq_graph_left,
                    0)

            dd_node_info_left = torch.matmul(dd_node_weight, dd_node_info_left)
            qd_node_info_left = torch.matmul(qd_node_weight, qd_node_info_left)
            qq_node_info_left = torch.matmul(qq_node_weight, qq_node_info_left)
            dq_node_info_left = torch.matmul(dq_node_weight, dq_node_info_left)


            dd_node_info_right = self._dd_node_fc_right(d_node)
            qd_node_info_right = self._qd_node_fc_right(d_node)
            qq_node_info_right = self._qq_node_fc_right(q_node)
            dq_node_info_right = self._dq_node_fc_right(q_node)

            dd_node_weight = util.replace_masked_values(
                    d_node_weight.unsqueeze(1).expand(-1, d_node_len, -1),
                    dd_graph_right,
                    0)

            qd_node_weight = util.replace_masked_values(
                    d_node_weight.unsqueeze(1).expand(-1, q_node_len, -1),
                    qd_graph_right,
                    0)

            qq_node_weight = util.replace_masked_values(
                    q_node_weight.unsqueeze(1).expand(-1, q_node_len, -1),
                    qq_graph_right,
                    0)

            dq_node_weight = util.replace_masked_values(
                    q_node_weight.unsqueeze(1).expand(-1, d_node_len, -1),
                    dq_graph_right,
                    0)

            dd_node_info_right = torch.matmul(dd_node_weight, dd_node_info_right)
            qd_node_info_right = torch.matmul(qd_node_weight, qd_node_info_right)
            qq_node_info_right = torch.matmul(qq_node_weight, qq_node_info_right)
            dq_node_info_right = torch.matmul(dq_node_weight, dq_node_info_right)


            agg_d_node_info = (dd_node_info_left + dd_node_info_right + dq_node_info_left + dq_node_info_right) / d_node_neighbor_num.unsqueeze(-1)
            agg_q_node_info = (qq_node_info_left + qq_node_info_right + qd_node_info_left + qd_node_info_right) / q_node_neighbor_num.unsqueeze(-1)

            d_node = F.relu(self_d_node_info + agg_d_node_info)
            q_node = F.relu(self_q_node_info + agg_q_node_info)


        all_d_weight = [weight.unsqueeze(1) for weight in all_d_weight]
        all_q_weight = [weight.unsqueeze(1) for weight in all_q_weight]

        all_d_weight = torch.cat(all_d_weight, dim=1)
        all_q_weight = torch.cat(all_q_weight, dim=1)

        return d_node, q_node, all_d_weight, all_q_weight # d_node_weight, q_node_weight


from dataclasses import dataclass, field
import random
from typing import Union, List

import numpy as np
import torch
from transformers import HfArgumentParser


@dataclass
class Arguments:
    output_dir: str = field(
        metadata={"help": "The output directory where the checkpoints will be written."},
    )
    data_path: str = field(
        metadata={"help": "The path of mmqa dataset."}
    )
    log_file: str = field (
        default='classifier.log',
        metadata= {"help": "The name of log"}
    )
    bert_model: str = field(
        default="microsoft/deberta-v3-large",
        metadata={"help": "The path of huggingface PLM"}
    )
    ckpt_file: str = field(
        default="classifier.pt",
        metadata= {"help": "The name of saved checkpoint"}
    )
    ckpt_for_test: str = field(
        default=None,
        metadata={"help": "if test_path is not None, load the ckpt from ckpt_for_test and predict the results"}
    )
    num_classes: int = field(
        default=4,
        metadata={"help": "The number of classes of labels"}
    )
    test_out_path: str = field(
        default="./prediction.json",
        metadata={"help": "The path of writing prediction"}
    )
    per_device_train_batch_size: int = field(
        default=16,
        metadata={"help": "If the train_mode set to sample, the per_device_train_batch_size must set to 1, "
                          "else if the train_mode set to row, you can set the per_device_train_batch_size to larger"}
    )
    per_device_eval_batch_size: int = field(
        default=1,
        metadata={"help": "one is the best"}
    )
    dataloader_pin_memory: bool = field(
        default=False,
        metadata={"help": "Whether to enable memory pinning for DataLoader. "
                          "Setting this to True can speed up data transfer "
                          "between CPU and GPU during training by pinning "
                          "the loaded data to CPU memory."}
    )
    dataloader_num_workers: int = field(
        default=0,
        metadata={"help": "The number of workers for DataLoader. '0' means using the main process only."}
    )
    dropout: float = field(default=0.2, metadata={"help": "Dropout of the classify head"})
    random_seed: int = field(default=42, metadata={"help": "Fix the random seed"})
    use_wandb: bool = field(default=True, metadata={"help": "Whether to use wandb for logging"})
    # truncation: int = field(default=128, metadata={"help": "Maximum sequence length for text truncation"})
    # eval_steps: int = field(default=10000, metadata={"help": "Evaluate the model every 'eval_steps' training steps"})
    train_mode: str = field(default="sample", metadata={"help": "Training mode. Possible values: 'sample', 'row'"})
    use_sampler: bool = field(default=False, metadata={"help": "Whether to use sampler for training data"})
    loss_fn_type: str = field(default="CE", metadata={"help": "Loss function type. Possible values: 'CE', 'BCE'"})
    learning_rate: float = field(default=7e-6, metadata={"help": "Learning rate for the optimizer"})
    num_train_epochs: int = field(default=1, metadata={"help": "Total number of training epochs"})
    # lr_scheduler_type: str = field(default="linear", metadata={
    #     "help": "Learning rate scheduler type. Possible values: 'linear', 'cosine', etc."})
    n_gpu: int = field(default=1, metadata={"help": "Total number of gpus"})
    adam_epsilon: float = field(default=1e-8,metadata={"help": "epsilon param of adam optimizer"})
    max_length: int = field(default=512, metadata={"help": "Max sequence length of bert model"})
    warmup_steps: int = field(default=1000, metadata={"help": "Number of warmup steps for the learning rate scheduler"})
    # logging_steps: int = field(default=100, metadata={"help": "Log training information every 'logging_steps' steps"})
    # save_steps: int = field(default=10000, metadata={"help": "Save the model every 'save_steps' steps"})
    resume_path: str = field(default=None, metadata={"help": "resume from checkpoint"})


def fix_seed(seed):
    # random
    random.seed(seed)
    # Numpy
    np.random.seed(seed)
    # Pytorch
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def get_training_args():
    return HfArgumentParser(
        Arguments
    ).parse_args_into_dataclasses()[0]


def get_recall(outputs, labels, top_n):
    if len(outputs.shape) > 1:
        outputs = outputs[:, 1]
    outputs, index = torch.sort(outputs, dim=0, descending=True)
    labels = labels[index]
    outputs[:top_n] = 1
    outputs[top_n:] = 0

    TP = ((outputs == 1).int() + (labels == 1).int()) == 2
    FN = ((outputs == 0).int() + (labels == 1).int()) == 2
    SE = float(TP.sum()) / (float((TP + FN).sum()) + 1e-6)
    return SE


def right_pad_sequences(sequences: List[list], padding_value: Union[int, bool] = 0,
                        max_len: int = -1, truncation: int = 512) -> torch.Tensor:
    max_len = max_len if max_len > 0 else max(len(s) for s in sequences)
    max_len = min(max_len, truncation)
    padded_seqs = []
    for seq in sequences:
        if len(seq) > truncation:
            seq = seq[:truncation]
        padded_seqs.append(
            torch.cat((torch.LongTensor(seq), (torch.full((max_len - len(seq),), padding_value, dtype=torch.long)))))
    return torch.stack(padded_seqs).cuda()


class Stack(object):
    def __init__(self):
        self.list = []
    def isEmpty(self):
        return self.list == []
    def push(self, item):
        self.list.append(item)
    def pop(self):
        return self.list.pop()
    def top(self):
        return self.list[len(self.list)-1]
    def size(self):
        return len(self.list)

def pretomid(x):
    s = Stack()
    list = x.split()
    for par in list:
        if par in "+-*/":
            s.push(par)
        else: 
            if s.top() in '+-*/':
                s.push(par)
            else:
                while (not s.isEmpty()) and (not s.top() in '+-*/'):
                    shu = s.pop()
                    fu = s.pop()
                    par = '( '+shu+fu+par+' )'
                s.push(str(par))

    answer = s.pop()
    answer = answer.replace('(','').replace(')','').replace('+',' + ').replace('*',' * ')
    return answer#



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch optimization for BERT model."""

import math
import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required
from torch.nn.utils import clip_grad_norm_

def warmup_cosine(x, warmup=0.002):
    if x < warmup:
        return x/warmup
    return 0.5 * (1.0 + torch.cos(math.pi * x))

def warmup_constant(x, warmup=0.002):
    if x < warmup:
        return x/warmup
    return 1.0

def warmup_linear(x, warmup=0.002):
    if x < warmup:
        return x/warmup
    return 1.0 - x

SCHEDULES = {
    'warmup_cosine':warmup_cosine,
    'warmup_constant':warmup_constant,
    'warmup_linear':warmup_linear,
}


class BertAdam(Optimizer):
    """Implements BERT version of Adam algorithm with weight decay fix.
    Params:
        lr: learning rate
        warmup: portion of t_total for the warmup, -1  means no warmup. Default: -1
        t_total: total number of training steps for the learning
            rate schedule, -1  means constant learning rate. Default: -1
        schedule: schedule to use for the warmup (see above). Default: 'warmup_linear'
        b1: Adams b1. Default: 0.9
        b2: Adams b2. Default: 0.999
        e: Adams epsilon. Default: 1e-6
        weight_decay: Weight decay. Default: 0.01
        max_grad_norm: Maximum norm for the gradients (-1 means no clipping). Default: 1.0
    """
    def __init__(self, params, lr=required, warmup=-1, t_total=-1, schedule='warmup_linear',
                 b1=0.9, b2=0.999, e=1e-6, weight_decay=0.01,
                 max_grad_norm=1.0):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {} - should be >= 0.0".format(lr))
        if schedule not in SCHEDULES:
            raise ValueError("Invalid schedule parameter: {}".format(schedule))
        if not 0.0 <= warmup < 1.0 and not warmup == -1:
            raise ValueError("Invalid warmup: {} - should be in [0.0, 1.0[ or -1".format(warmup))
        if not 0.0 <= b1 < 1.0:
            raise ValueError("Invalid b1 parameter: {} - should be in [0.0, 1.0[".format(b1))
        if not 0.0 <= b2 < 1.0:
            raise ValueError("Invalid b2 parameter: {} - should be in [0.0, 1.0[".format(b2))
        if not e >= 0.0:
            raise ValueError("Invalid epsilon value: {} - should be >= 0.0".format(e))
        defaults = dict(lr=lr, schedule=schedule, warmup=warmup, t_total=t_total,
                        b1=b1, b2=b2, e=e, weight_decay=weight_decay,
                        max_grad_norm=max_grad_norm)
        super(BertAdam, self).__init__(params, defaults)

    def get_lr(self):
        lr = []
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                if len(state) == 0:
                    return [0]
                if group['t_total'] != -1:
                    schedule_fct = SCHEDULES[group['schedule']]
                    lr_scheduled = group['lr'] * schedule_fct(state['step']/group['t_total'], group['warmup'])
                else:
                    lr_scheduled = group['lr']
                lr.append(lr_scheduled)
        return lr

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['next_m'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['next_v'] = torch.zeros_like(p.data)

                next_m, next_v = state['next_m'], state['next_v']
                beta1, beta2 = group['b1'], group['b2']

                # Add grad clipping
                if group['max_grad_norm'] > 0:
                    clip_grad_norm_(p, group['max_grad_norm'])

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                # next_m.mul_(beta1).add_(1 - beta1, grad)
                # import pdb; pdb.set_trace()
                next_m.mul_(beta1).add_(other=grad, alpha=1-beta1)
                next_v.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                update = next_m / (next_v.sqrt() + group['e'])

                # Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                if group['weight_decay'] > 0.0:
                    update += group['weight_decay'] * p.data

                if group['t_total'] != -1:
                    schedule_fct = SCHEDULES[group['schedule']]
                    lr_scheduled = group['lr'] * schedule_fct(state['step']/group['t_total'], group['warmup'])
                else:
                    lr_scheduled = group['lr']

                update_with_lr = lr_scheduled * update
                p.data.add_(-update_with_lr)

                state['step'] += 1

                # step_size = lr_scheduled * math.sqrt(bias_correction2) / bias_correction1
                # No bias correction
                # bias_correction1 = 1 - beta1 ** state['step']
                # bias_correction2 = 1 - beta2 ** state['step']

        return loss



import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import numpy as np
from transformers import BartTokenizer, BartForQuestionAnswering, BartForConditionalGeneration
from transformers import AutoTokenizer, RobertaTokenizer, RobertaModel, BertTokenizer, BertModel, BertForQuestionAnswering
import json
from tqdm import tqdm
import random
from pathlib import Path
import logging
import os, sys, math
import torch.nn.functional as F
from transformers import get_linear_schedule_with_warmup, AdamW
import argparse



class BartReadDataset(Dataset):
    def __init__(self, args, input_data, which_dataset):
        super(BartReadDataset, self).__init__()
        tokenizer = AutoTokenizer.from_pretrained(args.plm)
        self.tokenizer = tokenizer
        self.total_data = []
        for data in tqdm(input_data):
            path = args.WikiTables
            table_id = data['table_id']
            with open('{}/tables_tok/{}.json'.format(path, table_id), 'r') as f:
                table = json.load(f)  
            with open('{}/request_tok/{}.json'.format(path, table_id), 'r') as f:
                requested_document = json.load(f)
            answer = data['answer-text']
            
            if which_dataset == 'train':
                if len(data['row_gold']) == 0:
                    gold_row = np.argmax(np.array(data['row_pre_logit']))
                elif len(data['row_gold']) == 1:
                    gold_row = data['row_gold'][0]
                else:
                    logit = data['row_pre_logit']
                    index = np.argmax(np.array(logit)[data['row_gold']])
                    gold_row = data['row_gold'][index]
            elif which_dataset == 'dev':
                gold_row = np.argmax(np.array(data['row_pre_logit']))
                
            question_ids = self.tokenizer.encode(data['question'])

            headers = [_[0] for _ in table['header']]
            template = '{} is {}. '
            
            gold_row_data = table['data'][gold_row]
            
            links = []
            row_ids = []

            for j, cell in enumerate(gold_row_data):
                
                cell_string = template.format(headers[j], cell[0]) #cell[1]代表超链接
                cell_tokens = tokenizer.tokenize(cell_string)
                cell_ids = tokenizer.convert_tokens_to_ids(cell_tokens)
                row_ids.extend(cell_ids)
                row_ids += [tokenizer.sep_token_id] #加分隔符sep
                links.extend(cell[1])
            
            passage_ids = []
            for link in links:
                
                link_tokens = tokenizer.tokenize(requested_document[link]) #超链接
                link_ids = tokenizer.convert_tokens_to_ids(link_tokens)
                passage_ids.extend(link_ids)
                passage_ids += [tokenizer.sep_token_id]
                
            input_ids = question_ids + row_ids + passage_ids
            data['input_ids'] = input_ids
            
            answer_ids = self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(data['answer-text']))
            data['answer_ids'] = [self.tokenizer.eos_token_id] + answer_ids + [self.tokenizer.eos_token_id]

            self.total_data.append(data)
            
    def __len__(self):
        return len(self.total_data)
    def __getitem__(self, index):
        data = self.total_data[index]
        return data
        
        

class ReadDataset(Dataset):
    def __init__(self, args, data, which_dataset):
        super(ReadDataset, self).__init__()
        tokenizer = AutoTokenizer.from_pretrained(args.plm)
        self.tokenizer = tokenizer
        self.total_data = data
        total_data = []
        for data in tqdm(self.total_data):
            path = args.WikiTables
            table_id = data['table_id']
            with open('{}/tables_tok/{}.json'.format(path, table_id), 'r') as f:
                table = json.load(f)  
            with open('{}/request_tok/{}.json'.format(path, table_id), 'r') as f:
                requested_document = json.load(f)

            headers = [_[0] for _ in table['header']]
            row_tmp = '{} is {}. '
            
            answer = data['answer-text']
            if which_dataset == 'train':
                if len(data['row_gold']) == 0:
                    gold_row = np.argmax(np.array(data['row_pre_logit']))
                elif len(data['row_gold']) == 1:
                    gold_row = data['row_gold'][0]
                else:
                    logit = data['row_pre_logit']
                    index = np.argmax(np.array(logit)[data['row_gold']])
                    gold_row = data['row_gold'][index]
            elif which_dataset == 'dev':
                gold_row = np.argmax(np.array(data['row_pre_logit']))
                
            row = table['data'][gold_row]
            answer_ids = self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(answer))
            if which_dataset == 'train':
                question_ids = self.tokenizer.encode(data['question'])
                input_ids = []
                row_ids = []
                links = []
                for j, cell in enumerate(row):
                    if cell[0] != '':
                        cell_desc = row_tmp.format(headers[j], cell[0])
                        cell_toks = self.tokenizer.tokenize(cell_desc)
                        cell_ids = self.tokenizer.convert_tokens_to_ids(cell_toks)
                        row_ids += cell_ids
                    if cell[1] != []:
                        links += cell[1]
                links_ids = []

                for link in links:
                    passage_toks = self.tokenizer.tokenize(requested_document[link])
                    passage_ids = self.tokenizer.convert_tokens_to_ids(passage_toks)
                    links_ids += passage_ids
                input_ids = question_ids + row_ids + links_ids + [self.tokenizer.sep_token_id]
            else:
                input_ids, row_ids, links = [], [], []
                tokens = []
                tokens_lens = []
                ids_token_index = []

                tokens.append(tokenizer.cls_token)
                input_ids.append(tokenizer.cls_token_id)
                tokens_lens.append(1)

                question_tokens, question_input_ids, question_tokens_lens = self.string_tokenizer(tokenizer, data['question'])
                tokens += question_tokens
                input_ids += question_input_ids
                tokens_lens += question_tokens_lens


                tokens += [tokenizer.sep_token]
                input_ids += [tokenizer.sep_token_id]
                tokens_lens += [1]
                row_tokens, row_input_ids, row_tokens_lens, links = [], [], [], []
                for j, cell in enumerate(row):
                    if cell[0] != '':
                        cell_desc = row_tmp.format(headers[j], cell[0])
                        cell_tokens, cell_input_ids, cell_tokens_lens = self.string_tokenizer(tokenizer, cell_desc)
                        row_tokens += cell_tokens
                        row_input_ids += cell_input_ids
                        row_tokens_lens += cell_tokens_lens
                    if cell[1] != []:
                        links += cell[1]
                
                link_tokens, link_input_ids, link_tokens_lens = [], [], []

                for link in links:
                    passage = requested_document[link]
                    passage_tokens, passage_input_ids, passage_tokens_lens = self.string_tokenizer(tokenizer, passage)
                    link_tokens += passage_tokens
                    link_input_ids += passage_input_ids
                    link_tokens_lens += passage_tokens_lens
                
                input_ids += row_input_ids + link_input_ids + [tokenizer.sep_token_id]
                tokens += row_tokens + link_tokens + [tokenizer.sep_token]
                tokens_lens += row_tokens_lens + link_tokens_lens + [1]
                for i, item in enumerate(tokens_lens):
                    ids_token_index += [i] * item
                data['tokens'] = tokens
                data['ids_token_index'] = ids_token_index
            keneng_start = np.where(np.array(input_ids)==answer_ids[0])[0].tolist()
            start, end = 0, 0
            for keneng_start_i in keneng_start:
                if input_ids[keneng_start_i:keneng_start_i+len(answer_ids)] == answer_ids:
                    start = keneng_start_i
                    end = keneng_start_i+len(answer_ids) - 1
                    break
            if which_dataset == 'train':
                if input_ids[start:end+1] != answer_ids:
                    continue
                if start > 511 or end > 511:
                    continue
            data['start'] = start
            data['end'] = end
            data['input_ids'] = input_ids
            data['metadata'] = data
            total_data.append(data)
        self.data = total_data
        
    def string_tokenizer(self, tokenizer, input_string):
        str_tokens = input_string.split(' ')
        str_input_ids = []
        str_tokens_lens = []
        for i, token in enumerate(str_tokens):
            str_input_ids_i = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(token))
            str_input_ids.extend(str_input_ids_i)
            str_tokens_lens.extend([len(str_input_ids_i)])
        return str_tokens, str_input_ids, str_tokens_lens

    def __len__(self):
        return len(self.data)
    def __getitem__(self, index):
        data = self.data[index]
        return data
        





from fcntl import F_GETLEASE
import re
import string
from typing import List
import numpy as np

def scale_to_num(scale):
    scale = scale.lower()
    num = 1
    if 'hundred' in scale:  # hundred
        num = 100
    elif 'thousand' in scale:  # thousand
        num = 1000
    elif 'million' in scale:  # million
        num = 1000000
    elif 'billion' in scale:  # billion
        num = 1000000000
    elif 'percent' in scale:  # percent
        num = 0.01
    return num

def extract_one_num_from_str(s):
    s = _clean_num(s)
    r_num = r"([+-]?\d+(\.\d+)?)|([+-]?\.\d+)"
    groups = re.findall(r_num, s)
    if len(groups) == 0:
        return None
    num = groups[0][0]
    if num == '':
        return None
    if '.' in num:
        return float(num)
    return int(num)

EXCLUDE_IN_NUM = "'\"\\$€£¥%(),[]"
def _clean_num(text:str):
    return "".join([ch for ch in str(text) if ch not in EXCLUDE_IN_NUM])


def is_year(item):
    years = [str(i) for i in range(2001, 2051)]
    for year in years:
        if year in item:
            return True
    return False



def is_number(text: str) -> bool:
    if is_year(text):
        return False
    
    # if ' - ' in text:
    #     return True
        
        
    text = text.replace('£','').replace('m','').replace('b','').replace('bn','').replace('years', '').replace('–', '-').replace('RMB', '').replace('HKD', '').replace('S', '').replace(';','').replace('year','').replace('(in','').replace('€m','')
    if text != '' and text[-1] == '.':
        text = text[:-1]
    try:
        words = " ".join([_clean_num(w) for w in text.split()]).split()
        if len(words) == 0:
            """1023 or 1 million"""
            return False
        num = float(words[0])
        if np.isnan(num):
            return False
        if len(words) >= 2:
            if scale_to_num(words[1]) == 1:
                return False
        return True
    except ValueError:
        return False
    # except AttributeError:
    #     return False

def is_number_q(text: str) -> bool:
        
        
    text = text.replace('£','').replace('m','').replace('b','').replace('bn','').replace('years', '').replace('–', '-').replace('RMB', '').replace('HKD', '').replace('S', '').replace(';','')
    if text != '' and text[-1] == '.':
        text = text[:-1]
    try:
        words = " ".join([_clean_num(w) for w in text.split()]).split()
        if len(words) == 0:
            """1023 or 1 million"""
            return False
        num = float(words[0])
        if np.isnan(num):
            return False
        if len(words) >= 2:
            if scale_to_num(words[1]) == 1:
                return False
        return True
    except ValueError:
        return False


def negative_num_handle(x):
    """
    :param x:  transform (134) -> -134
    :return:
    """
    x = x.replace(',','')
    x = x.replace('%','')
    all = re.findall('(\([\d.\s]+\))', x.strip())
    if len(all) > 0:
        return -1
    return 1

def percent_num_handle(x):
    """
    :param x:  transform 12% -> 12/100
    :return:
    """
    if '%' in x:
        all = 1
    else:
        all = 0
    if all:
        return 0.01
    return 1

def word_scale_handle(x):
    """
    :param x: 1 million = 1,000,000
    :return:
    """
    iter = re.finditer('([\d.]+\s?[a-zA-Z]+)', x)
    for one in iter:
        text = one.group(0).lower()
        scale_val = scale_to_num(text)
        return scale_val
    return 1

def to_number(text:str) -> float:

    # if ' - ' in text:
    #     return text.split('-')[0].strip()
    try:
        if text[0] == '.' and is_number(text.replace('.','0.')) and is_number(text.replace('.','')):
            text = text.replace('.','0.')
    except:
        import pdb; pdb.set_trace()
    text = text.replace("–", '-').replace('years', '').replace('$', '').replace('year','').replace('(in','').replace('€m','')
    num = extract_one_num_from_str(text)
    scale_val = word_scale_handle(text)
    negative_flag = negative_num_handle(text)
    percent_flag = percent_num_handle(text)
    if num is not None:
        return round(num * scale_val * negative_flag * percent_flag, 4)
    return None

def remove_articles(text: str) -> str:
    regex = re.compile(r'\b(a|an|the)\b', re.UNICODE)
    return re.sub(regex, ' ', text)

def white_space_fix(text: str) -> str:
    return ' '.join(text.split())

EXCLUDE = set(string.punctuation)
def remove_punc(text: str) -> str:
    if not is_number(text):
        return ''.join(ch for ch in text if ch not in EXCLUDE)
    else:
        return text

def lower(text: str) -> str:
    return text.lower()

def tokenize(text: str) -> List[str]:
    return re.split(" ", text)


def normalize_number(text: str) -> str:
    if is_number(text):
        return str(to_number(text))
    else:
        return text

def normalize_answer(text: str) -> str:
    """Lower text and remove punctuation, articles and extra whitespace."""
    parts = [white_space_fix(remove_articles(normalize_number(remove_punc(lower(token)))))
             for token in tokenize(text)]
    parts = [part for part in parts if part.strip()]
    normalized = ' '.join(parts).strip()
    return normalized


STRIPPED_CHARACTERS = string.punctuation + ''.join([u"‘", u"’", u"´", u"`", "_"])
def ws_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip().lower()
    if not text:
        return []
    text = white_space_fix(text)
    tokens = text.split()
    tokens = [token.strip(STRIPPED_CHARACTERS) for token in tokens]
    return tokens




import json
import argparse

from Retrieval import RetrievalTrainer

def read_config(path):
    with open(path, 'r') as file:
        config = json.load(file)
    return config



if __name__ == '__main__':
    parser  = argparse.ArgumentParser()
    parser.add_argument('--train_data_path', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/train_r_col.json')
    parser.add_argument('--dev_data_path', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/dev_r_col.json')
    parser.add_argument('--WikiTables', type=str, default='/home/lfy/S3HQA1/Data/HybridQA/WikiTables-WithLinks')
    parser.add_argument('--is_train', type=int, default=1)
    parser.add_argument('--output_path', type=str, default='./try')
    parser.add_argument('--plm', type=str, default='/home/lfy/S3HQA1/PTM/bert-base-uncased')
    parser.add_argument('--mode', choices=['row', 'column'])
    parser.add_argument('--config_file', type=str, default='config_retriever.json')
    
    args = parser.parse_args()
    
    training_config = read_config(args.config_file)
    
    mytrainer = RetrievalTrainer(args)
    mytrainer.train(**training_config)
    



import torch
import torch.nn as nn
from tatqa_metric import TaTQAEmAndF1
from .tools.util import FFNLayer
from .tools import allennlp as util
from typing import Dict, List, Tuple
import numpy as np
from reg_hnt.data.file_utils import is_scatter_available
import math
from reg_hnt.reghnt.model_utils import rnn_wrapper, lens2mask, PoolingFunction, FFN
from reg_hnt.reghnt.new_model import Arithmetic, SubwordAggregation, InputRNNLayer, RGTAT, OtherModel
from reg_hnt.reghnt.modeling_tree import Prediction
from reg_hnt.reghnt.Vocabulary import *
from reg_hnt.data.tatqa_dataset import is_year, clean_year

np.set_printoptions(threshold=np.inf)
if is_scatter_available():
    from torch_scatter import scatter
    from torch_scatter import scatter_max


class RegHNTModel(nn.Module):
    def __init__(self,
                 encoder,
                 config,
                 bsz,
                 operator_classes: int,
                 scale_classes: int,
                 operator_criterion: nn.CrossEntropyLoss = None,
                 scale_criterion: nn.CrossEntropyLoss = None,
                 hidden_size: int = None,
                 dropout_prob: float = None,
                 arithmetic_op_index: List = None,
                 op_mode: int = None,
                 ablation_mode: int = None,
                 ):
        super(RegHNTModel, self).__init__()
        self.encoder = encoder
        self.config = config
        self.operator_classes = operator_classes
        self.scale_classes = scale_classes
        if hidden_size is None:
            hidden_size = self.config.hidden_size
            self.hidden_size = hidden_size
        if dropout_prob is None:
            dropout_prob = self.config.hidden_dropout_prob

        self.operator_predictor = FFNLayer(hidden_size, hidden_size, operator_classes, dropout_prob)
        
        self.scale_predictor = FFNLayer(3*hidden_size, hidden_size, scale_classes, dropout_prob)
        
        self.dropout_layer = nn.Dropout(p=0.2)

        self.subword_aggregation = SubwordAggregation(hidden_size, subword_aggregation="attentive-pooling")

        self.operator_criterion = operator_criterion
        self.scale_criterion = scale_criterion
        self.config = config
        self.rnn_layer = InputRNNLayer(hidden_size, hidden_size, cell='lstm', schema_aggregation='attentive-pooling')
        self.hidden_layer = RGTAT(gnn_hidden_size=hidden_size, gnn_num_layers=8, \
            num_heads=8, relation_num=len(Relation_vocabulary['word2id']))
        
        self.arithmetic_model = Arithmetic(hidden_size)
        self.other_model = OtherModel(hidden_size)
        
        self.arithmetic_op_index = arithmetic_op_index

        self._metrics = TaTQAEmAndF1()
        

    def forward(self,**batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        token_type_ids = batch['token_type_ids']
        operator_class = batch['operator_class']
        question_ids = batch['question_id']
        position_ids = None
        device = input_ids.device
        
        outputs = self.encoder(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids)
        
        sequence_output = outputs[0]
        batch_size = sequence_output.size(0)

        
        question, table, paragraph = self.subword_aggregation(sequence_output, batch)
        input_dict = {
            "question": self.dropout_layer(question),
            "table": self.dropout_layer(table),
            "paragraph": self.dropout_layer(paragraph)
        }
        
        inputs = self.rnn_layer(input_dict, batch)

        inputs = self.hidden_layer(inputs, batch)
        
        
        hidden = inputs.new_zeros(batch_size, batch['mask']['b'].size(1), self.hidden_size)
        hidden_output = hidden.masked_scatter_(batch['mask']['b'].unsqueeze(-1), inputs)

        question_hidden = util.replace_masked_values(hidden_output, batch['mask']['question'].unsqueeze(-1).long(), 0)
        question_mean = torch.mean(question_hidden, dim=1)
        table_hidden = util.replace_masked_values(hidden_output, batch['mask']['table'].unsqueeze(-1).long(), 0)
        table_mean = torch.mean(table_hidden, dim=1)
        paragraph_hidden = util.replace_masked_values(hidden_output, batch['mask']['paragraph'].unsqueeze(-1).long(), 0)
        paragraph_mean = torch.mean(paragraph_hidden, dim=1)
        
        q_t_p_mean = torch.cat((question_mean, table_mean, paragraph_mean), dim=-1)
        scale_prediction = self.scale_predictor(q_t_p_mean)

        operator_prediction = self.operator_predictor(sequence_output[:, 0])
        
        outputs = batch['outputs']
        operator_class = batch['operator_class'].tolist()
        
        #####################################################################################################
        arithmetic_labels = list(map(lambda i: 1 if operator_class[i]==OPERATOR_CLASSES["ARITHMETIC"] else 0, range(len(operator_class))))
        arithmetic_num = arithmetic_labels.count(1)
        other_labels = list([1 for i in range(batch_size)])
        
        sequence_output_ari, hidden_output_ari = [], []
        
        for i in range(batch_size):
            if arithmetic_labels[i] == 1:
                sequence_output_ari.append(sequence_output[i])
                hidden_output_ari.append(hidden_output[i])
        

        if arithmetic_num > 0:
            sequence_output_ari = torch.stack(sequence_output_ari, dim=0)
            hidden_output_ari = torch.stack(hidden_output_ari, dim=0)
        
        ######################################################################################################
        arithmetic_loss, other_loss = 0, 0
        table_tag_prediction, paragraph_tag_prediction = None, None
        if arithmetic_num > 0:
            arithmetic_loss, all_node_outputs, target = self.arithmetic_model(sequence_output_ari, hidden_output_ari, batch, arithmetic_labels)
        
        other_loss, table_tag_prediction, paragraph_tag_prediction = self.other_model(sequence_output, hidden_output, batch, other_labels)

        operator_prediction_loss = self.operator_criterion(operator_prediction, torch.tensor(operator_class).cuda())
        
        scale_class = batch['scale_class']
        scale_prediction_loss = self.scale_criterion(scale_prediction, scale_class)

        output_dict = {}
        output_dict['loss'] = arithmetic_loss + other_loss + operator_prediction_loss + scale_prediction_loss
        output_dict["question_id"] = []
        output_dict["answer"] = []
        output_dict["scale"] = []

        LOss = output_dict["loss"].item()
        if np.isnan(LOss):
            import pdb; pdb.set_trace()

        with torch.no_grad():
 
            predicted_operator_class = torch.argmax(operator_prediction, dim=-1).tolist()
            predicted_scale_class = torch.argmax(scale_prediction, dim=-1).detach().cpu().numpy()
            
            for i in range(len(predicted_operator_class)):
                if predicted_operator_class[i] == OPERATOR_CLASSES["ARITHMETIC"]:
                    sequence_output_i = sequence_output[i].unsqueeze(0)
                    hidden_output_i = hidden_output[i].unsqueeze(0)
                    batch_label_i = list(map(lambda x: 1 if x==i else 0, range(batch_size)))
                    answer = self.arithmetic_model.predict(sequence_output_i, hidden_output_i, batch, batch_label_i, int(predicted_scale_class[i]), batch['answer_dict'][i])
                else:
                    sequence_output_i = sequence_output[i].unsqueeze(0)
                    hidden_output_i = hidden_output[i].unsqueeze(0)
                    batch_label_i = list(map(lambda x: 1 if x==i else 0, range(batch_size)))
                    answer = self.other_model.predict(hidden_output_i, batch, batch_label_i, predicted_operator_class[i], table_tag_prediction[i].unsqueeze(0), paragraph_tag_prediction[i].unsqueeze(0))

                if not answer:
                    answer = -1
                
                output_dict["answer"].append(answer)
                output_dict["scale"].append(SCALE[int(predicted_scale_class[i])])
                output_dict["question_id"].append(question_ids[i])
                
                self._metrics(batch['answer_dict'][i], answer, SCALE[int(predicted_scale_class[i])])

        return output_dict
    
    def predict(self, **batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        token_type_ids = batch['token_type_ids']
        question_ids = batch['question_id']
        position_ids = None

        outputs = self.encoder(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids)
        
        sequence_output = outputs[0]
        batch_size = sequence_output.size(0)
        
        question, table, paragraph = self.subword_aggregation(sequence_output, batch)
        input_dict = {
            "question": self.dropout_layer(question),
            "table": self.dropout_layer(table),
            "paragraph": self.dropout_layer(paragraph)
        }
        
        inputs = self.rnn_layer(input_dict, batch)
        
        inputs = self.hidden_layer(inputs, batch)
        
        hidden = inputs.new_zeros(batch_size, batch['mask']['b'].size(1), self.hidden_size)
        hidden_output = hidden.masked_scatter_(batch['mask']['b'].unsqueeze(-1), inputs)

        question_hidden = util.replace_masked_values(hidden_output, batch['mask']['question'].unsqueeze(-1).long(), 0)
        question_mean = torch.mean(question_hidden, dim=1)
        table_hidden = util.replace_masked_values(hidden_output, batch['mask']['table'].unsqueeze(-1).long(), 0)
        table_mean = torch.mean(table_hidden, dim=1)
        paragraph_hidden = util.replace_masked_values(hidden_output, batch['mask']['paragraph'].unsqueeze(-1).long(), 0)
        paragraph_mean = torch.mean(paragraph_hidden, dim=1)
        
        q_t_p_mean = torch.cat((question_mean, table_mean, paragraph_mean), dim=-1)
        scale_prediction = self.scale_predictor(q_t_p_mean)
        operator_prediction = self.operator_predictor(sequence_output[:, 0])
        
        predicted_operator_class = torch.argmax(operator_prediction, dim=-1).tolist()
        predicted_scale_class = torch.argmax(scale_prediction, dim=-1).detach().cpu().numpy()

        output_dict = {}
        
        output_dict["question_id"] = []
        output_dict["answer"] = []
        output_dict["scale"] = []
        output_dict["gold_answers"] = []
        answer = -1
        if predicted_operator_class[0] == OPERATOR_CLASSES["ARITHMETIC"]:
            answer = self.arithmetic_model.predict(sequence_output, hidden_output, batch, [1], int(predicted_scale_class[0]), batch['answer_dict'][0])
        else:
            _, table_tag_prediction, paragraph_tag_prediction = self.other_model.forward2(sequence_output, hidden_output, batch, [1])
            answer = self.other_model.predict(hidden_output, batch, [1], predicted_operator_class[0], table_tag_prediction, paragraph_tag_prediction)


        if "which year " in batch['questions'][0] or "Which year " in batch['questions'][0] or "Which FY " in batch['questions'][0] or "which FY " in batch['questions'][0]:
            if isinstance(answer, list) and len(answer) == 1 and is_year(answer[0]):
                answer[0] = clean_year(answer[0])
        
        if "which years" in batch['questions'][0] or "Which years" in batch['questions'][0]:
            if isinstance(answer, list) and len(answer) > 1:
                for i in range(len(answer)):
                    if is_year(answer[i]):
                        answer[i] = clean_year(answer[i])

        
        output_dict["answer"].append(answer)
        output_dict["scale"].append(SCALE[int(predicted_scale_class[0])])
        output_dict["question_id"].append(question_ids[0])
        output_dict["gold_answers"].append(batch['answer_dict'][0])

        current_op = OPERATOR_CLASSES_R[predicted_operator_class[0]]
        gold_op = OPERATOR_CLASSES_R[batch['operator_class'].tolist()[0]]

        
        self._metrics({**batch['answer_dict'][0], "uid": question_ids[0]}, answer,
                SCALE[int(predicted_scale_class[0])], None, None, 
                pred_op=current_op, gold_op=gold_op)
        return output_dict
        
        
    def predict2(self, **batch):
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        token_type_ids = batch['token_type_ids']
        question_ids = batch['question_id']
        position_ids = None

        outputs = self.encoder(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids)
        
        sequence_output = outputs[0]
        batch_size = sequence_output.size(0)
        
        question, table, paragraph = self.subword_aggregation(sequence_output, batch)
        input_dict = {
            "question": self.dropout_layer(question),
            "table": self.dropout_layer(table),
            "paragraph": self.dropout_layer(paragraph)
        }
        
        inputs = self.rnn_layer(input_dict, batch)
        
        inputs = self.hidden_layer(inputs, batch)
        
        hidden = inputs.new_zeros(batch_size, batch['mask']['b'].size(1), self.hidden_size)
        hidden_output = hidden.masked_scatter_(batch['mask']['b'].unsqueeze(-1), inputs)

        question_hidden = util.replace_masked_values(hidden_output, batch['mask']['question'].unsqueeze(-1).long(), 0)
        question_mean = torch.mean(question_hidden, dim=1)
        table_hidden = util.replace_masked_values(hidden_output, batch['mask']['table'].unsqueeze(-1).long(), 0)
        table_mean = torch.mean(table_hidden, dim=1)
        paragraph_hidden = util.replace_masked_values(hidden_output, batch['mask']['paragraph'].unsqueeze(-1).long(), 0)
        paragraph_mean = torch.mean(paragraph_hidden, dim=1)
        
        q_t_p_mean = torch.cat((question_mean, table_mean, paragraph_mean), dim=-1)
        scale_prediction = self.scale_predictor(q_t_p_mean)
        operator_prediction = self.operator_predictor(sequence_output[:, 0])
        
        
        
        predicted_operator_class = torch.argmax(operator_prediction, dim=-1).tolist()
        predicted_scale_class = torch.argmax(scale_prediction, dim=-1).detach().cpu().numpy()

        output_dict = {}
        
        output_dict["question_id"] = []
        output_dict["answer"] = []
        output_dict["scale"] = []
        output_dict["operator"] = []

        if predicted_operator_class[0] == OPERATOR_CLASSES["ARITHMETIC"]:
            answer = self.arithmetic_model.predict(sequence_output, hidden_output, batch, [1], int(predicted_scale_class[0]), 0)
        else:
            _, table_tag_prediction, paragraph_tag_prediction = self.other_model.forward2(sequence_output, hidden_output, batch, [1])
            answer = self.other_model.predict(hidden_output, batch, [1], predicted_operator_class[0], table_tag_prediction, paragraph_tag_prediction)


        if "which year " in batch['questions'][0] or "Which year " in batch['questions'][0] or "Which FY " in batch['questions'][0] or "which FY " in batch['questions'][0]:
            if isinstance(answer, list) and len(answer) == 1 and is_year(answer[0]):
                answer[0] = clean_year(answer[0])
        
        if "which years" in batch['questions'][0] or "Which years" in batch['questions'][0]:
            if isinstance(answer, list) and len(answer) > 1:
                for i in range(len(answer)):
                    if is_year(answer[i]):
                        answer[i] = clean_year(answer[i])

        
        output_dict["answer"].append(answer)
        output_dict["scale"].append(SCALE[int(predicted_scale_class[0])])
        output_dict["question_id"].append(question_ids[0])
        output_dict["operator"].append(predicted_operator_class[0])

        

        return output_dict
        
        
    def reset(self):
        self._metrics.reset()

    def set_metrics_mdoe(self, mode):
        self._metrics = TaTQAEmAndF1(mode=mode)

    def get_metrics(self, logger=None, reset: bool = False) -> Dict[str, float]:
        detail_em, detail_f1 = self._metrics.get_detail_metric()

        raw_detail = self._metrics.get_raw_pivot_table()
        exact_match, f1_score, scale_score, op_score = self._metrics.get_overall_metric(reset)
        print(f"raw matrix:{raw_detail}\r\n")
        print(f"detail em:{detail_em}\r\n")
        print(f"detail f1:{detail_f1}\r\n")
        print(f"global em:{exact_match}\r\n")
        print(f"global f1:{f1_score}\r\n")
        print(f"global scale:{scale_score}\r\n")
        print(f"global op:{op_score}\r\n")
        if logger is not None:
            logger.info(f"raw matrix:{raw_detail}\r\n")
            logger.info(f"detail em:{detail_em}\r\n")
            logger.info(f"detail f1:{detail_f1}\r\n")
            logger.info(f"global em:{exact_match}\r\n")
            logger.info(f"global f1:{f1_score}\r\n")
            logger.info(f"global scale:{scale_score}\r\n")
            logger.info(f"global op:{op_score}\r\n")
        return {'em': exact_match, 'f1': f1_score, "scale": scale_score, "op": op_score}

    def get_df(self):
        raws = self._metrics.get_raw()
        detail_em, detail_f1 = self._metrics.get_detail_metric()
        raw_detail = self._metrics.get_raw_pivot_table()
        return detail_em, detail_f1, raws, raw_detail


    
